# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Tools for comparing two :py:class:`~libbe.bugdir.BugDir`\s.
"""

import difflib
import types

import libbe
import libbe.bugdir
import libbe.bug
import libbe.util.tree
from libbe.storage.util.settings_object import setting_name_to_attr_name
from libbe.util.utility import time_to_str


class SubscriptionType (libbe.util.tree.Tree):
    """Trees of subscription types to allow users to select exactly what
    notifications they want to subscribe to.
    """
    def __init__(self, type_name, *args, **kwargs):
        libbe.util.tree.Tree.__init__(self, *args, **kwargs)
        self.type = type_name
    def __str__(self):
        return self.type
    def __cmp__(self, other):
        return cmp(self.type, other.type)
    def __repr__(self):
        return '<SubscriptionType: %s>' % str(self)
    def string_tree(self, indent=0):
        lines = []
        for depth,node in self.thread():
            lines.append('%s%s' % (' '*(indent+2*depth), node))
        return '\n'.join(lines)

BUGDIR_ID = 'DIR'
BUGDIR_TYPE_NEW = SubscriptionType('new')
BUGDIR_TYPE_MOD = SubscriptionType('mod')
BUGDIR_TYPE_REM = SubscriptionType('rem')
BUGDIR_TYPE_ALL = SubscriptionType('all',
                      [BUGDIR_TYPE_NEW, BUGDIR_TYPE_MOD, BUGDIR_TYPE_REM])

# same name as BUGDIR_TYPE_ALL for consistency
BUG_TYPE_ALL = SubscriptionType(str(BUGDIR_TYPE_ALL))

INVALID_TYPE = SubscriptionType('INVALID')

class InvalidType (ValueError):
    def __init__(self, type_name, type_root):
        msg = 'Invalid type %s for tree:\n%s' \
            % (type_name, type_root.string_tree(4))
        ValueError.__init__(self, msg)
        self.type_name = type_name
        self.type_root = type_root

def type_from_name(name, type_root, default=None, default_ok=False):
    if name == str(type_root):
        return type_root
    for t in type_root.traverse():
        if name == str(t):
            return t
    if default_ok:
        return default
    raise InvalidType(name, type_root)

class Subscription (object):
    """A user subscription.

    Examples
    --------

    >>> subscriptions = [Subscription('XYZ', 'all'),
    ...                  Subscription('DIR', 'new'),
    ...                  Subscription('ABC', BUG_TYPE_ALL),]
    >>> print sorted(subscriptions)
    [<Subscription: DIR (new)>, <Subscription: ABC (all)>, <Subscription: XYZ (all)>]
    """
    def __init__(self, id, subscription_type, **kwargs):
        if 'type_root' not in kwargs:
            if id == BUGDIR_ID:
                kwargs['type_root'] = BUGDIR_TYPE_ALL
            else:
                kwargs['type_root'] = BUG_TYPE_ALL
        if type(subscription_type) in types.StringTypes:
            subscription_type = type_from_name(subscription_type, **kwargs)
        self.id = id
        self.type = subscription_type
    def __cmp__(self, other):
        for attr in 'id', 'type':
            value = cmp(getattr(self, attr), getattr(other, attr))
            if value != 0:
                if self.id == BUGDIR_ID:
                    return -1
                elif other.id == BUGDIR_ID:
                    return 1
                return value
    def __str__(self):
        return str(self.type)
    def __repr__(self):
        return '<Subscription: %s (%s)>' % (self.id, self.type)

def subscriptions_from_string(string=None, subscription_sep=',', id_sep=':'):
    """Provide a simple way for non-Python interfaces to read in subscriptions.

    Examples
    --------

    >>> subscriptions_from_string(None)
    [<Subscription: DIR (all)>]
    >>> subscriptions_from_string('DIR:new,DIR:rem,ABC:all,XYZ:all')
    [<Subscription: DIR (new)>, <Subscription: DIR (rem)>, <Subscription: ABC (all)>, <Subscription: XYZ (all)>]
    >>> subscriptions_from_string('DIR::new')
    Traceback (most recent call last):
      ...
    ValueError: Invalid subscription "DIR::new", should be ID:TYPE
    """
    if string == None:
        return [Subscription(BUGDIR_ID, BUGDIR_TYPE_ALL)]
    subscriptions = []
    for subscription in string.split(','):
        fields = subscription.split(':')
        if len(fields) != 2:
            raise ValueError('Invalid subscription "%s", should be ID:TYPE'
                             % subscription)
        id,type = fields
        subscriptions.append(Subscription(id, type))
    return subscriptions

class DiffTree (libbe.util.tree.Tree):
    """A tree holding difference data for easy report generation.

    Examples
    --------

    >>> bugdir = DiffTree('bugdir')
    >>> bdsettings = DiffTree('settings', data='target: None -> 1.0')
    >>> bugdir.append(bdsettings)
    >>> bugs = DiffTree('bugs', 'bug-count: 5 -> 6')
    >>> bugdir.append(bugs)
    >>> new = DiffTree('new', 'new bugs: ABC, DEF')
    >>> bugs.append(new)
    >>> rem = DiffTree('rem', 'removed bugs: RST, UVW')
    >>> bugs.append(rem)
    >>> print bugdir.report_string()
    target: None -> 1.0
    bug-count: 5 -> 6
      new bugs: ABC, DEF
      removed bugs: RST, UVW
    >>> print '\\n'.join(bugdir.paths())
    bugdir
    bugdir/settings
    bugdir/bugs
    bugdir/bugs/new
    bugdir/bugs/rem
    >>> bugdir.child_by_path('/') == bugdir
    True
    >>> bugdir.child_by_path('/bugs') == bugs
    True
    >>> bugdir.child_by_path('/bugs/rem') == rem
    True
    >>> bugdir.child_by_path('bugdir') == bugdir
    True
    >>> bugdir.child_by_path('bugdir/') == bugdir
    True
    >>> bugdir.child_by_path('bugdir/bugs') == bugs
    True
    >>> bugdir.child_by_path('/bugs').masked = True
    >>> print bugdir.report_string()
    target: None -> 1.0
    """
    def __init__(self, name, data=None, data_part_fn=str,
                 requires_children=False, masked=False):
        libbe.util.tree.Tree.__init__(self)
        self.name = name
        self.data = data
        self.data_part_fn = data_part_fn
        self.requires_children = requires_children
        self.masked = masked
    def paths(self, parent_path=None):
        paths = []
        if parent_path == None:
            path = self.name
        else:
            path = '%s/%s' % (parent_path, self.name)
        paths.append(path)
        for child in self:
            paths.extend(child.paths(path))
        return paths
    def child_by_path(self, path):
        if hasattr(path, 'split'): # convert string path to a list of names
            names = path.split('/')
            if names[0] == '':
                names[0] = self.name # replace root with self
            if len(names) > 1 and names[-1] == '':
                names = names[:-1] # strip empty tail
        else: # it was already an array
            names = path
        assert len(names) > 0, path
        if names[0] == self.name:
            if len(names) == 1:
                return self
            for child in self:
                if names[1] == child.name:
                    return child.child_by_path(names[1:])
        if len(names) == 1:
            raise KeyError, "%s doesn't match '%s'" % (names, self.name)
        raise KeyError, '%s points to child not in %s' % (names, [c.name for c in self])
    def report_string(self):
        report = self.report()
        if report == None:
            return ''
        return '\n'.join(report)
    def report(self, root=None, parent=None, depth=0):
        if root == None:
            root = self.make_root()
        if self.masked == True:
            return root
        data_part = self.data_part(depth)
        if self.requires_children == True \
                and len([c for c in self if c.masked == False]) == 0:
            pass
        else:
            self.join(root, parent, data_part)
            if data_part != None:
                depth += 1
            for child in self:
                root = child.report(root, self, depth)
        return root
    def make_root(self):
        return []
    def join(self, root, parent, data_part):
        if data_part != None:
            root.append(data_part)
    def data_part(self, depth, indent=True):
        if self.data == None:
            return None
        if hasattr(self, '_cached_data_part'):
            return self._cached_data_part
        data_part = self.data_part_fn(self.data)
        if indent == True:
            data_part_lines = data_part.splitlines()
            indent = '  '*(depth)
            line_sep = '\n'+indent
            data_part = indent+line_sep.join(data_part_lines)
        self._cached_data_part = data_part
        return data_part

class Diff (object):
    """Difference tree generator for BugDirs.

    Examples
    --------

    >>> import copy
    >>> bd = libbe.bugdir.SimpleBugDir(memory=True)
    >>> bd_new = copy.deepcopy(bd)
    >>> bd_new.target = '1.0'
    >>> a = bd_new.bug_from_uuid('a')
    >>> rep = a.comment_root.new_reply("I'm closing this bug")
    >>> rep.uuid = 'acom'
    >>> rep.author = 'John Doe <j@doe.com>'
    >>> rep.date = 'Thu, 01 Jan 1970 00:00:00 +0000'
    >>> a.status = 'closed'
    >>> b = bd_new.bug_from_uuid('b')
    >>> bd_new.remove_bug(b)
    >>> c = bd_new.new_bug('Bug C', _uuid='c')
    >>> d = Diff(bd, bd_new)
    >>> r = d.report_tree()
    >>> print '\\n'.join(r.paths())
    bugdir
    bugdir/settings
    bugdir/bugs
    bugdir/bugs/new
    bugdir/bugs/new/c
    bugdir/bugs/rem
    bugdir/bugs/rem/b
    bugdir/bugs/mod
    bugdir/bugs/mod/a
    bugdir/bugs/mod/a/settings
    bugdir/bugs/mod/a/comments
    bugdir/bugs/mod/a/comments/new
    bugdir/bugs/mod/a/comments/new/acom
    bugdir/bugs/mod/a/comments/rem
    bugdir/bugs/mod/a/comments/mod
    >>> print r.report_string()
    Changed bug directory settings:
      target: None -> 1.0
    New bugs:
      abc/c:om: Bug C
    Removed bugs:
      abc/b:cm: Bug B
    Modified bugs:
      abc/a:cm: Bug A
        Changed bug settings:
          status: open -> closed
        New comments:
          from John Doe <j@doe.com> on Thu, 01 Jan 1970 00:00:00 +0000
            I'm closing this bug...

    You can also limit the report generation by providing a list of
    subscriptions.

    >>> subscriptions = [Subscription('DIR', BUGDIR_TYPE_NEW),
    ...                  Subscription('b', BUG_TYPE_ALL)]
    >>> r = d.report_tree(subscriptions)
    >>> print r.report_string()
    New bugs:
      abc/c:om: Bug C
    Removed bugs:
      abc/b:cm: Bug B

    While sending subscriptions to report_tree() makes the report
    generation more efficient (because you may not need to compare
    _all_ the bugs, etc.), sometimes you will have several sets of
    subscriptions.  In that case, it's better to run full_report()
    first, and then use report_tree() to avoid redundant comparisons.

    >>> d.full_report()
    >>> print d.report_tree([subscriptions[0]]).report_string()
    New bugs:
      abc/c:om: Bug C
    >>> print d.report_tree([subscriptions[1]]).report_string()
    Removed bugs:
      abc/b:cm: Bug B

    >>> bd.cleanup()
    """
    def __init__(self, old_bugdir, new_bugdir):
        self.old_bugdir = old_bugdir
        self.new_bugdir = new_bugdir

    # data assembly methods

    def _changed_bugs(self, subscriptions):
        """
        Search for differences in all bugs between .old_bugdir and
        .new_bugdir.  Returns
          (added_bugs, modified_bugs, removed_bugs)
        where added_bugs and removed_bugs are lists of added and
        removed bugs respectively.  modified_bugs is a list of
        (old_bug,new_bug) pairs.
        """
        bugdir_types = [s.type for s in subscriptions if s.id == BUGDIR_ID]
        new_uuids = []
        old_uuids = []
        for bd_type in [BUGDIR_TYPE_ALL, BUGDIR_TYPE_NEW, BUGDIR_TYPE_MOD]:
            if bd_type in bugdir_types:
                new_uuids = list(self.new_bugdir.uuids())
                break
        for bd_type in [BUGDIR_TYPE_ALL, BUGDIR_TYPE_REM]:
            if bd_type in bugdir_types:
                old_uuids = list(self.old_bugdir.uuids())
                break
        subscribed_bugs = []
        for s in subscriptions:
            if s.id != BUGDIR_ID:
                try:
                    bug = self.new_bugdir.bug_from_uuid(s.id)
                except libbe.bugdir.NoBugMatches:
                    bug = self.old_bugdir.bug_from_uuid(s.id)
                subscribed_bugs.append(bug.uuid)
        new_uuids.extend([s for s in subscribed_bugs
                          if self.new_bugdir.has_bug(s)])
        new_uuids = sorted(set(new_uuids))
        old_uuids.extend([s for s in subscribed_bugs
                          if self.old_bugdir.has_bug(s)])
        old_uuids = sorted(set(old_uuids))

        added = []
        removed = []
        modified = []
        if hasattr(self.old_bugdir, 'changed'):
            # take advantage of a RevisionedBugDir-style changed() method
            new_ids,mod_ids,rem_ids = self.old_bugdir.changed()
            for id in new_ids:
                for a_id in self.new_bugdir.storage.ancestors(id):
                    if a_id.count('/') == 0:
                        if a_id in [b.id.storage() for b in added]:
                            break
                        try:
                            bug = self.new_bugdir.bug_from_uuid(a_id)
                            added.append(bug)
                        except libbe.bugdir.NoBugMatches:
                            pass
            for id in rem_ids:
                for a_id in self.old_bugdir.storage.ancestors(id):
                    if a_id.count('/') == 0:
                        if a_id in [b.id.storage() for b in removed]:
                            break
                        try:
                            bug = self.old_bugdir.bug_from_uuid(a_id)
                            removed.append(bug)
                        except libbe.bugdir.NoBugMatches:
                            pass
            for id in mod_ids:
                for a_id in self.new_bugdir.storage.ancestors(id):
                    if a_id.count('/') == 0:
                        if a_id in [b[0].id.storage() for b in modified]:
                            break
                        try:
                            new_bug = self.new_bugdir.bug_from_uuid(a_id)
                            old_bug = self.old_bugdir.bug_from_uuid(a_id)
                            modified.append((old_bug, new_bug))
                        except libbe.bugdir.NoBugMatches:
                            pass
        else:
            for uuid in new_uuids:
                new_bug = self.new_bugdir.bug_from_uuid(uuid)
                try:
                    old_bug = self.old_bugdir.bug_from_uuid(uuid)
                except KeyError:
                    if BUGDIR_TYPE_ALL in bugdir_types \
                            or BUGDIR_TYPE_NEW in bugdir_types \
                            or uuid in subscribed_bugs:
                        added.append(new_bug)
                    continue
                if BUGDIR_TYPE_ALL in bugdir_types \
                        or BUGDIR_TYPE_MOD in bugdir_types \
                        or uuid in subscribed_bugs:
                    if old_bug.storage != None and old_bug.storage.is_readable():
                        old_bug.load_comments()
                    if new_bug.storage != None and new_bug.storage.is_readable():
                        new_bug.load_comments()
                    if old_bug != new_bug:
                        modified.append((old_bug, new_bug))
            for uuid in old_uuids:
                if not self.new_bugdir.has_bug(uuid):
                    old_bug = self.old_bugdir.bug_from_uuid(uuid)
                    removed.append(old_bug)
        added.sort()
        removed.sort()
        modified.sort(self._bug_modified_cmp)
        return (added, modified, removed)
    def _bug_modified_cmp(self, left, right):
        return cmp(left[1], right[1])
    def _changed_comments(self, old, new):
        """
        Search for differences in all loaded comments between the bugs
        old and new.  Returns
          (added_comments, modified_comments, removed_comments)
        analogous to ._changed_bugs.
        """
        if hasattr(self, '__changed_comments'):
            if new.uuid in self.__changed_comments:
                return self.__changed_comments[new.uuid]
        else:
            self.__changed_comments = {}
        added = []
        removed = []
        modified = []
        old.comment_root.sort(key=lambda comm : comm.time)
        new.comment_root.sort(key=lambda comm : comm.time)
        old_comment_ids = [c.uuid for c in old.comments()]
        new_comment_ids = [c.uuid for c in new.comments()]
        for uuid in new_comment_ids:
            new_comment = new.comment_from_uuid(uuid)
            try:
                old_comment = old.comment_from_uuid(uuid)
            except KeyError:
                added.append(new_comment)
            else:
                if old_comment != new_comment:
                    modified.append((old_comment, new_comment))
        for uuid in old_comment_ids:
            if uuid not in new_comment_ids:
                old_comment = old.comment_from_uuid(uuid)
                removed.append(old_comment)
        self.__changed_comments[new.uuid] = (added, modified, removed)
        return self.__changed_comments[new.uuid]
    def _attribute_changes(self, old, new, attributes):
        """
        Take two objects old and new, and compare the value of *.attr
        for attr in the list attribute names.  Returns a list of
          (attr_name, old_value, new_value)
        tuples.
        """
        change_list = []
        for attr in attributes:
            old_value = getattr(old, attr)
            new_value = getattr(new, attr)
            if old_value != new_value:
                change_list.append((attr, old_value, new_value))
        if len(change_list) >= 0:
            return change_list
        return None
    def _settings_properties_attribute_changes(self, old, new,
                                              hidden_properties=[]):
        properties = sorted(new.settings_properties)
        for p in hidden_properties:
            properties.remove(p)
        attributes = [setting_name_to_attr_name(None, p)
                      for p in properties]
        return self._attribute_changes(old, new, attributes)
    def _bugdir_attribute_changes(self):
        return self._settings_properties_attribute_changes( \
            self.old_bugdir, self.new_bugdir)
    def _bug_attribute_changes(self, old, new):
        return self._settings_properties_attribute_changes(old, new)
    def _comment_attribute_changes(self, old, new):
        return self._settings_properties_attribute_changes(old, new)

    # report generation methods

    def full_report(self, diff_tree=DiffTree):
        """
        Generate a full report for efficiency if you'll be using
        .report_tree() with several sets of subscriptions.
        """
        self._cached_full_report = self.report_tree(diff_tree=diff_tree,
                                                    allow_cached=False)
        self._cached_full_report_diff_tree = diff_tree
    def _sub_report(self, subscriptions):
        """
        Return ._cached_full_report masked for subscriptions.
        """
        root = self._cached_full_report
        bugdir_types = [s.type for s in subscriptions if s.id == BUGDIR_ID]
        subscribed_bugs = [s.id for s in subscriptions
                           if BUG_TYPE_ALL.has_descendant( \
                                     s.type, match_self=True)]
        selected_by_bug = [node.name
                           for node in root.child_by_path('bugdir/bugs')]
        if BUGDIR_TYPE_ALL in bugdir_types:
            for node in root.traverse():
                node.masked = False
            selected_by_bug = []
        else:
            try:
                node = root.child_by_path('bugdir/settings')
                node.masked = True
            except KeyError:
                pass
        for name,type in (('new', BUGDIR_TYPE_NEW),
                          ('mod', BUGDIR_TYPE_MOD),
                          ('rem', BUGDIR_TYPE_REM)):
            if type in bugdir_types:
                bugs = root.child_by_path('bugdir/bugs/%s' % name)
                for bug_node in bugs:
                    for node in bug_node.traverse():
                        node.masked = False
                selected_by_bug.remove(name)
        for name in selected_by_bug:
            bugs = root.child_by_path('bugdir/bugs/%s' % name)
            for bug_node in bugs:
                if bug_node.name in subscribed_bugs:
                    for node in bug_node.traverse():
                        node.masked = False
                else:
                    for node in bug_node.traverse():
                        node.masked = True
        return root
    def report_tree(self, subscriptions=None, diff_tree=DiffTree,
                    allow_cached=True):
        """
        Pretty bare to make it easy to adjust to specific cases.  You
        can pass in a DiffTree subclass via diff_tree to override the
        default report assembly process.
        """
        if allow_cached == True \
                and hasattr(self, '_cached_full_report') \
                and diff_tree == self._cached_full_report_diff_tree:
            return self._sub_report(subscriptions)
        if subscriptions == None:
            subscriptions = [Subscription(BUGDIR_ID, BUGDIR_TYPE_ALL)]
        bugdir_settings = sorted(self.new_bugdir.settings_properties)
        root = diff_tree('bugdir')
        bugdir_subscriptions = [s.type for s in subscriptions
                                if s.id == BUGDIR_ID]
        if BUGDIR_TYPE_ALL in bugdir_subscriptions:
            bugdir_attribute_changes = self._bugdir_attribute_changes()
            if len(bugdir_attribute_changes) > 0:
                bugdir = diff_tree('settings', bugdir_attribute_changes,
                                   self.bugdir_attribute_change_string)
                root.append(bugdir)
        bug_root = diff_tree('bugs')
        root.append(bug_root)
        add,mod,rem = self._changed_bugs(subscriptions)
        bnew = diff_tree('new', 'New bugs:', requires_children=True)
        bug_root.append(bnew)
        for bug in add:
            b = diff_tree(bug.uuid, bug, self.bug_add_string)
            bnew.append(b)
        brem = diff_tree('rem', 'Removed bugs:', requires_children=True)
        bug_root.append(brem)
        for bug in rem:
            b = diff_tree(bug.uuid, bug, self.bug_rem_string)
            brem.append(b)
        bmod = diff_tree('mod', 'Modified bugs:', requires_children=True)
        bug_root.append(bmod)
        for old,new in mod:
            b = diff_tree(new.uuid, (old,new), self.bug_mod_string)
            bmod.append(b)
            bug_attribute_changes = self._bug_attribute_changes(old, new)
            if len(bug_attribute_changes) > 0:
                bset = diff_tree('settings', bug_attribute_changes,
                                 self.bug_attribute_change_string)
                b.append(bset)
            if old.summary != new.summary:
                data = (old.summary, new.summary)
                bsum = diff_tree('summary', data, self.bug_summary_change_string)
                b.append(bsum)
            cr = diff_tree('comments')
            b.append(cr)
            a,m,d = self._changed_comments(old, new)
            cnew = diff_tree('new', 'New comments:', requires_children=True)
            for comment in a:
                c = diff_tree(comment.uuid, comment, self.comment_add_string)
                cnew.append(c)
            crem = diff_tree('rem', 'Removed comments:',requires_children=True)
            for comment in d:
                c = diff_tree(comment.uuid, comment, self.comment_rem_string)
                crem.append(c)
            cmod = diff_tree('mod','Modified comments:',requires_children=True)
            for o,n in m:
                c = diff_tree(n.uuid, (o,n), self.comment_mod_string)
                cmod.append(c)
                comm_attribute_changes = self._comment_attribute_changes(o, n)
                if len(comm_attribute_changes) > 0:
                    cset = diff_tree('settings', comm_attribute_changes,
                                     self.comment_attribute_change_string)
                if o.body != n.body:
                    data = (o.body, n.body)
                    cbody = diff_tree('cbody', data,
                                      self.comment_body_change_string)
                    c.append(cbody)
            cr.extend([cnew, crem, cmod])
        return root

    # change data -> string methods.
    # Feel free to play with these in subclasses.

    def attribute_change_string(self, attribute_changes, indent=0):
        indent_string = '  '*indent
        change_strings = [u'%s: %s -> %s' % f for f in attribute_changes]
        for i,change_string in enumerate(change_strings):
            change_strings[i] = indent_string+change_string
        return u'\n'.join(change_strings)
    def bugdir_attribute_change_string(self, attribute_changes):
        return 'Changed bug directory settings:\n%s' % \
            self.attribute_change_string(attribute_changes, indent=1)
    def bug_attribute_change_string(self, attribute_changes):
        return 'Changed bug settings:\n%s' % \
            self.attribute_change_string(attribute_changes, indent=1)
    def comment_attribute_change_string(self, attribute_changes):
        return 'Changed comment settings:\n%s' % \
            self.attribute_change_string(attribute_changes, indent=1)
    def bug_add_string(self, bug):
        return bug.string(shortlist=True)
    def bug_rem_string(self, bug):
        return bug.string(shortlist=True)
    def bug_mod_string(self, bugs):
        old_bug,new_bug = bugs
        return new_bug.string(shortlist=True)
    def bug_summary_change_string(self, summaries):
        old_summary,new_summary = summaries
        return 'summary changed:\n  %s\n  %s' % (old_summary, new_summary)
    def _comment_summary_string(self, comment):
        return 'from %s on %s' % (comment.author, time_to_str(comment.time))
    def comment_add_string(self, comment):
        summary = self._comment_summary_string(comment)
        first_line = comment.body.splitlines()[0]
        return '%s\n  %s...' % (summary, first_line)
    def comment_rem_string(self, comment):
        summary = self._comment_summary_string(comment)
        first_line = comment.body.splitlines()[0]
        return '%s\n  %s...' % (summary, first_line)
    def comment_mod_string(self, comments):
        old_comment,new_comment = comments
        return self._comment_summary_string(new_comment)
    def comment_body_change_string(self, bodies):
        old_body,new_body = bodies
        return ''.join(difflib.unified_diff(
                old_body.splitlines(True),
                new_body.splitlines(True),
                'before', 'after'))



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Alexander Belchenko <bialix@ukr.net>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Oleg Romanyshyn <oromanyshyn@panoramicfeedback.com>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Define :py:class:`BugDir` for storing a collection of bugs.
"""

import copy
import errno
import os
import os.path
import time
import types
try: # import core module, Python >= 2.5
    from xml.etree import ElementTree
except ImportError: # look for non-core module
    from elementtree import ElementTree
import xml.sax.saxutils

import libbe
import libbe.storage as storage
from libbe.storage.util.properties import Property, doc_property, \
    local_property, defaulting_property, checked_property, \
    fn_checked_property, cached_property, primed_property, \
    change_hook_property, settings_property
import libbe.storage.util.settings_object as settings_object
import libbe.storage.util.mapfile as mapfile
import libbe.bug as bug
import libbe.util.utility as utility
import libbe.util.id

if libbe.TESTING == True:
    import doctest
    import sys
    import unittest

    import libbe.storage.base


class NoBugMatches(libbe.util.id.NoIDMatches):
    def __init__(self, *args, **kwargs):
        libbe.util.id.NoIDMatches.__init__(self, *args, **kwargs)
    def __str__(self):
        if self.msg == None:
            return 'No bug matches %s' % self.id
        return self.msg


class BugDir (list, settings_object.SavedSettingsObject):
    """A BugDir is a container for :py:class:`~libbe.bug.Bug`\s, with some
    additional attributes.

    Parameters
    ----------
    storage : :py:class:`~libbe.storage.base.Storage`
       Storage instance containing the bug directory.  If
       `from_storage` is `False`, `storage` may be `None`.
    uuid : str, optional
       Set the bugdir UUID (see :py:mod:`libbe.util.id`).
       Useful if you are loading one of several bugdirs
       stored in a single Storage instance.
    from_storage : bool, optional
       If `True`, attempt to load from storage.  Otherwise,
       setup in memory, saving to `storage` if it is not `None`.

    See Also
    --------
    SimpleBugDir : bugdir manipulation exampes.
    """

    settings_properties = []
    required_saved_properties = []
    _prop_save_settings = settings_object.prop_save_settings
    _prop_load_settings = settings_object.prop_load_settings
    def _versioned_property(settings_properties=settings_properties,
                            required_saved_properties=required_saved_properties,
                            **kwargs):
        if "settings_properties" not in kwargs:
            kwargs["settings_properties"] = settings_properties
        if "required_saved_properties" not in kwargs:
            kwargs["required_saved_properties"]=required_saved_properties
        return settings_object.versioned_property(**kwargs)

    @_versioned_property(name="target",
                         doc="The current project development target.")
    def target(): return {}

    def _setup_severities(self, severities):
        if severities not in [None, settings_object.EMPTY]:
            bug.load_severities(severities)
    def _set_severities(self, old_severities, new_severities):
        self._setup_severities(new_severities)
        self._prop_save_settings(old_severities, new_severities)
    @_versioned_property(name="severities",
                         doc="The allowed bug severities and their descriptions.",
                         change_hook=_set_severities)
    def severities(): return {}

    def _setup_status(self, active_status, inactive_status):
        bug.load_status(active_status, inactive_status)
    def _set_active_status(self, old_active_status, new_active_status):
        self._setup_status(new_active_status, self.inactive_status)
        self._prop_save_settings(old_active_status, new_active_status)
    @_versioned_property(name="active_status",
                         doc="The allowed active bug states and their descriptions.",
                         change_hook=_set_active_status)
    def active_status(): return {}

    def _set_inactive_status(self, old_inactive_status, new_inactive_status):
        self._setup_status(self.active_status, new_inactive_status)
        self._prop_save_settings(old_inactive_status, new_inactive_status)
    @_versioned_property(name="inactive_status",
                         doc="The allowed inactive bug states and their descriptions.",
                         change_hook=_set_inactive_status)
    def inactive_status(): return {}

    def _extra_strings_check_fn(value):
        return utility.iterable_full_of_strings(value, \
                         alternative=settings_object.EMPTY)
    def _extra_strings_change_hook(self, old, new):
        self.extra_strings.sort() # to make merging easier
        self._prop_save_settings(old, new)
    @_versioned_property(name="extra_strings",
                         doc="Space for an array of extra strings.  Useful for storing state for functionality implemented purely in becommands/<some_function>.py.",
                         default=[],
                         check_fn=_extra_strings_check_fn,
                         change_hook=_extra_strings_change_hook,
                         mutable=True)
    def extra_strings(): return {}

    def _bug_map_gen(self):
        map = {}
        for bug in self:
            map[bug.uuid] = bug
        for uuid in self.uuids():
            if uuid not in map:
                map[uuid] = None
        self._bug_map_value = map # ._bug_map_value used by @local_property

    @Property
    @primed_property(primer=_bug_map_gen)
    @local_property("bug_map")
    @doc_property(doc="A dict of (bug-uuid, bug-instance) pairs.")
    def _bug_map(): return {}

    def __init__(self, storage, uuid=None, from_storage=False):
        list.__init__(self)
        settings_object.SavedSettingsObject.__init__(self)
        self.storage = storage
        self.id = libbe.util.id.ID(self, 'bugdir')
        self.uuid = uuid
        if from_storage == True:
            if self.uuid == None:
                self.uuid = [c for c in self.storage.children()
                             if c != 'version'][0]
            self.load_settings()
        else:
            if self.uuid == None:
                self.uuid = libbe.util.id.uuid_gen()
            if self.storage != None and self.storage.is_writeable():
                self.save()

    # methods for saving/loading/accessing settings and properties.

    def load_settings(self, settings_mapfile=None):
        if settings_mapfile == None:
            settings_mapfile = \
                self.storage.get(self.id.storage('settings'), default='{}\n')
        try:
            settings = mapfile.parse(settings_mapfile)
        except mapfile.InvalidMapfileContents, e:
            raise Exception('Invalid settings file for bugdir %s\n'
                            '(BE version missmatch?)' % self.id.user())
        self._setup_saved_settings(settings)
        self._setup_severities(self.severities)
        self._setup_status(self.active_status, self.inactive_status)

    def save_settings(self):
        mf = mapfile.generate(self._get_saved_settings())
        self.storage.set(self.id.storage('settings'), mf)

    def load_all_bugs(self):
        """
        Warning: this could take a while.
        """
        self._clear_bugs()
        for uuid in self.uuids():
            self._load_bug(uuid)

    def save(self):
        """
        Save any loaded contents to storage.  Because of lazy loading
        of bugs and comments, this is actually not too inefficient.

        However, if self.storage.is_writeable() == True, then any
        changes are automatically written to storage as soon as they
        happen, so calling this method will just waste time (unless
        something else has been messing with your stored files).
        """
        self.storage.add(self.id.storage(), directory=True)
        self.storage.add(self.id.storage('settings'), parent=self.id.storage(),
                         directory=False)
        self.save_settings()
        for bug in self:
            bug.bugdir = self
            bug.storage = self.storage
            bug.save()

    # methods for managing bugs

    def uuids(self, use_cached_disk_uuids=True):
        if use_cached_disk_uuids==False or not hasattr(self, '_uuids_cache'):
            self._refresh_uuid_cache()
        self._uuids_cache = self._uuids_cache.union([bug.uuid for bug in self])
        return self._uuids_cache

    def _refresh_uuid_cache(self):
        self._uuids_cache = set()
        # list bugs that are in storage
        if self.storage != None and self.storage.is_readable():
            child_uuids = libbe.util.id.child_uuids(
                self.storage.children(self.id.storage()))
            for id in child_uuids:
                self._uuids_cache.add(id)

    def _clear_bugs(self):
        while len(self) > 0:
            self.pop()
        if hasattr(self, '_uuids_cache'):
            del(self._uuids_cache)
        self._bug_map_gen()

    def _load_bug(self, uuid):
        bg = bug.Bug(bugdir=self, uuid=uuid, from_storage=True)
        self.append(bg)
        self._bug_map_gen()
        return bg

    def new_bug(self, summary=None, _uuid=None):
        bg = bug.Bug(bugdir=self, uuid=_uuid, summary=summary,
                     from_storage=False)
        self.append(bg, update=True)
        return bg

    def append(self, bug, update=False):
        super(BugDir, self).append(bug)
        if update:
            bug.bugdir = self
            bug.storage = self.storage
            self._bug_map_gen()
            if (hasattr(self, '_uuids_cache') and
                not bug.uuid in self._uuids_cache):
                self._uuids_cache.add(bug.uuid)

    def remove_bug(self, bug):
        if hasattr(self, '_uuids_cache') and bug.uuid in self._uuids_cache:
            self._uuids_cache.remove(bug.uuid)
        self.remove(bug)
        if self.storage != None and self.storage.is_writeable():
            bug.remove()

    def bug_from_uuid(self, uuid):
        if not self.has_bug(uuid):
            raise NoBugMatches(
                uuid, self.uuids(),
                'No bug matches %s in %s' % (uuid, self.storage))
        if self._bug_map[uuid] == None:
            self._load_bug(uuid)
        return self._bug_map[uuid]

    def has_bug(self, bug_uuid):
        if bug_uuid not in self._bug_map:
            self._bug_map_gen()
            if bug_uuid not in self._bug_map:
                return False
        return True

    def xml(self, indent=0, show_bugs=False, show_comments=False):
        """
        >>> bug.load_severities(bug.severity_def)
        >>> bug.load_status(
        ...     active_status_def=bug.active_status_def,
        ...     inactive_status_def=bug.inactive_status_def)
        >>> bugdirA = SimpleBugDir(memory=True)
        >>> bugdirA.severities
        >>> bugdirA.severities = (('minor', 'The standard bug level.'),)
        >>> bugdirA.inactive_status = (
        ...     ('closed', 'The bug is no longer relevant.'),)
        >>> bugA = bugdirA.bug_from_uuid('a')
        >>> commA = bugA.comment_root.new_reply(body='comment A')
        >>> commA.uuid = 'commA'
        >>> commA.date = 'Thu, 01 Jan 1970 00:03:00 +0000'
        >>> print(bugdirA.xml(show_bugs=True, show_comments=True))
        ... # doctest: +REPORT_UDIFF
        <bugdir>
          <uuid>abc123</uuid>
          <short-name>abc</short-name>
          <severities>
            <entry>
              <key>minor</key>
              <value>The standard bug level.</value>
            </entry>
          </severities>
          <inactive-status>
            <entry>
              <key>closed</key>
              <value>The bug is no longer relevant.</value>
            </entry>
          </inactive-status>
          <bug>
            <uuid>a</uuid>
            <short-name>abc/a</short-name>
            <severity>minor</severity>
            <status>open</status>
            <creator>John Doe &lt;jdoe@example.com&gt;</creator>
            <created>Thu, 01 Jan 1970 00:00:00 +0000</created>
            <summary>Bug A</summary>
            <comment>
              <uuid>commA</uuid>
              <short-name>abc/a/com</short-name>
              <author></author>
              <date>Thu, 01 Jan 1970 00:03:00 +0000</date>
              <content-type>text/plain</content-type>
              <body>comment A</body>
            </comment>
          </bug>
          <bug>
            <uuid>b</uuid>
            <short-name>abc/b</short-name>
            <severity>minor</severity>
            <status>closed</status>
            <creator>Jane Doe &lt;jdoe@example.com&gt;</creator>
            <created>Thu, 01 Jan 1970 00:00:00 +0000</created>
            <summary>Bug B</summary>
          </bug>
        </bugdir>
        >>> bug.load_severities(bug.severity_def)
        >>> bug.load_status(
        ...     active_status_def=bug.active_status_def,
        ...     inactive_status_def=bug.inactive_status_def)
        >>> bugdirA.cleanup()
        """
        info = [('uuid', self.uuid),
                ('short-name', self.id.user()),
                ('target', self.target),
                ('severities', self.severities),
                ('active-status', self.active_status),
                ('inactive-status', self.inactive_status),
                ]
        lines = ['<bugdir>']
        for (k,v) in info:
            if v is not None:
                if k in ['severities', 'active-status', 'inactive-status']:
                    lines.append('  <{0}>'.format(k))
                    for vk,vv in v:
                        lines.extend([
                                '    <entry>',
                                '      <key>{0}</key>'.format(
                                    xml.sax.saxutils.escape(vk)),
                                '      <value>{0}</value>'.format(
                                    xml.sax.saxutils.escape(vv)),
                                '    </entry>',
                                ])
                    lines.append('  </{0}>'.format(k))
                else:
                    v = xml.sax.saxutils.escape(v)
                    lines.append('  <{0}>{1}</{0}>'.format(k, v))
        for estr in self.extra_strings:
            lines.append('  <extra-string>{0}</extra-string>'.format(estr))
        if show_bugs:
            for bug in self:
                bug_xml = bug.xml(indent=indent+2, show_comments=show_comments)
                if bug_xml:
                    bug_xml = bug_xml[indent:]  # strip leading indent spaces
                    lines.append(bug_xml)
        lines.append('</bugdir>')
        istring = ' '*indent
        sep = '\n' + istring
        return istring + sep.join(lines).rstrip('\n')

    def from_xml(self, xml_string, preserve_uuids=False, verbose=True):
        """
        Note: If a bugdir uuid is given, set .alt_id to it's value.
        >>> bug.load_severities(bug.severity_def)
        >>> bug.load_status(
        ...     active_status_def=bug.active_status_def,
        ...     inactive_status_def=bug.inactive_status_def)
        >>> bugdirA = SimpleBugDir(memory=True)
        >>> bugdirA.severities = (('minor', 'The standard bug level.'),)
        >>> bugdirA.inactive_status = (
        ...     ('closed', 'The bug is no longer relevant.'),)
        >>> bugA = bugdirA.bug_from_uuid('a')
        >>> commA = bugA.comment_root.new_reply(body='comment A')
        >>> commA.uuid = 'commA'
        >>> xml = bugdirA.xml(show_bugs=True, show_comments=True)
        >>> bugdirB = BugDir(storage=None)
        >>> bugdirB.from_xml(xml)
        >>> bugdirB.xml(show_bugs=True, show_comments=True) == xml
        False
        >>> bugdirB.uuid = bugdirB.alt_id
        >>> for bug_ in bugdirB:
        ...     bug_.uuid = bug_.alt_id
        ...     bug_.alt_id = None
        ...     for comm in bug_.comments():
        ...         comm.uuid = comm.alt_id
        ...         comm.alt_id = None
        >>> bugdirB.xml(show_bugs=True, show_comments=True) == xml
        True
        >>> bugdirB.explicit_attrs  # doctest: +NORMALIZE_WHITESPACE
        ['severities', 'inactive_status']
        >>> bugdirC = BugDir(storage=None)
        >>> bugdirC.from_xml(xml, preserve_uuids=True)
        >>> bugdirC.uuid == bugdirA.uuid
        True
        >>> bugdirC.xml(show_bugs=True, show_comments=True) == xml
        True
        >>> bug.load_severities(bug.severity_def)
        >>> bug.load_status(
        ...     active_status_def=bug.active_status_def,
        ...     inactive_status_def=bug.inactive_status_def)
        >>> bugdirA.cleanup()
        """
        if type(xml_string) == types.UnicodeType:
            xml_string = xml_string.strip().encode('unicode_escape')
        if hasattr(xml_string, 'getchildren'): # already an ElementTree Element
            bugdir = xml_string
        else:
            bugdir = ElementTree.XML(xml_string)
        if bugdir.tag != 'bugdir':
            raise utility.InvalidXML(
                'bugdir', bugdir, 'root element must be <bugdir>')
        tags = ['uuid', 'short-name', 'target', 'severities', 'active-status',
                'inactive-status', 'extra-string']
        self.explicit_attrs = []
        uuid = None
        estrs = []
        for child in bugdir.getchildren():
            if child.tag == 'short-name':
                pass
            elif child.tag == 'bug':
                bg = bug.Bug(bugdir=self)
                bg.from_xml(
                    child, preserve_uuids=preserve_uuids, verbose=verbose)
                self.append(bg, update=True)
                continue
            elif child.tag in tags:
                if child.text == None or len(child.text) == 0:
                    text = settings_object.EMPTY
                elif child.tag in ['severities', 'active-status',
                                   'inactive-status']:
                    entries = []
                    for entry in child.getchildren():
                        if entry.tag != 'entry':
                            raise utility.InvalidXML(
                                '{0} child element {1} must be <entry>'.format(
                                    child.tag, entry))
                        key = value = None
                        for kv in entry.getchildren():
                            if kv.tag == 'key':
                                if key is not None:
                                    raise utility.InvalidXML(
                                        ('duplicate keys ({0} and {1}) in {2}'
                                         ).format(key, kv.text, child.tag))
                                key = xml.sax.saxutils.unescape(kv.text)
                            elif kv.tag == 'value':
                                if value is not None:
                                    raise utility.InvalidXML(
                                        ('duplicate values ({0} and {1}) '
                                         'in {2}'
                                         ).format(
                                            value, kv.text, child.tag))
                                value = xml.sax.saxutils.unescape(kv.text)
                            else:
                                raise utility.InvalidXML(
                                    ('{0} child element {1} must be <key> or '
                                     '<value>').format(child.tag, kv))
                        if key is None:
                            raise utility.InvalidXML(
                                'no key for {0}'.format(child.tag))
                        if value is None:
                            raise utility.InvalidXML(
                                'no key for {0}'.format(child.tag))
                        entries.append((key, value))
                    text = entries
                else:
                    text = xml.sax.saxutils.unescape(child.text)
                    if not isinstance(text, unicode):
                        text = text.decode('unicode_escape')
                    text = text.strip()
                if child.tag == 'uuid' and not preserve_uuids:
                    uuid = text
                    continue # don't set the bug's uuid tag.
                elif child.tag == 'extra-string':
                    estrs.append(text)
                    continue # don't set the bug's extra_string yet.
                attr_name = child.tag.replace('-','_')
                self.explicit_attrs.append(attr_name)
                setattr(self, attr_name, text)
            elif verbose == True:
                sys.stderr.write('Ignoring unknown tag {0} in {1}\n'.format(
                        child.tag, bugdir.tag))
        if uuid != self.uuid:
            if not hasattr(self, 'alt_id') or self.alt_id == None:
                self.alt_id = uuid
        self.extra_strings = estrs

    def merge(self, other, accept_changes=True,
              accept_extra_strings=True, accept_bugs=True,
              accept_comments=True, change_exception=False):
        """Merge info from other into this bugdir.

        Overrides any attributes in self that are listed in
        other.explicit_attrs.

        >>> bugdirA = SimpleBugDir()
        >>> bugdirA.extra_strings += ['TAG: favorite']
        >>> bugdirB = SimpleBugDir()
        >>> bugdirB.explicit_attrs = ['target']
        >>> bugdirB.target = '1234'
        >>> bugdirB.extra_strings += ['TAG: very helpful']
        >>> bugdirB.extra_strings += ['TAG: useful']
        >>> bugA = bugdirB.bug_from_uuid('a')
        >>> commA = bugA.comment_root.new_reply(body='comment A')
        >>> commA.uuid = 'uuid-commA'
        >>> commA.date = 'Thu, 01 Jan 1970 00:01:00 +0000'
        >>> bugC = bugdirB.new_bug(summary='bug C', _uuid='c')
        >>> bugC.alt_id = 'alt-c'
        >>> bugC.time_string = 'Thu, 01 Jan 1970 00:02:00 +0000'
        >>> bugdirA.merge(
        ...     bugdirB, accept_changes=False, accept_extra_strings=False,
        ...     accept_bugs=False, change_exception=False)
        >>> print(bugdirA.target)
        None
        >>> bugdirA.merge(
        ...     bugdirB, accept_changes=False, accept_extra_strings=False,
        ...     accept_bugs=False, change_exception=True)
        Traceback (most recent call last):
          ...
        ValueError: Merge would change target "None"->"1234" for bugdir abc123
        >>> print(bugdirA.target)
        None
        >>> bugdirA.merge(
        ...     bugdirB, accept_changes=True, accept_extra_strings=False,
        ...     accept_bugs=False, change_exception=True)
        Traceback (most recent call last):
          ...
        ValueError: Merge would add extra string "TAG: useful" for bugdir abc123
        >>> print(bugdirA.target)
        1234
        >>> print(bugdirA.extra_strings)
        ['TAG: favorite']
        >>> bugdirA.merge(
        ...     bugdirB, accept_changes=True, accept_extra_strings=True,
        ...     accept_bugs=False, change_exception=True)
        Traceback (most recent call last):
          ...
        ValueError: Merge would add bug c (alt: alt-c) to bugdir abc123
        >>> print(bugdirA.extra_strings)
        ['TAG: favorite', 'TAG: useful', 'TAG: very helpful']
        >>> bugdirA.merge(
        ...     bugdirB, accept_changes=True, accept_extra_strings=True,
        ...     accept_bugs=True, change_exception=True)
        >>> print(bugdirA.xml(show_bugs=True, show_comments=True))
        ... # doctest: +ELLIPSIS, +REPORT_UDIFF
        <bugdir>
          <uuid>abc123</uuid>
          <short-name>abc</short-name>
          <target>1234</target>
          <extra-string>TAG: favorite</extra-string>
          <extra-string>TAG: useful</extra-string>
          <extra-string>TAG: very helpful</extra-string>
          <bug>
            <uuid>a</uuid>
            <short-name>abc/a</short-name>
            <severity>minor</severity>
            <status>open</status>
            <creator>John Doe &lt;jdoe@example.com&gt;</creator>
            <created>Thu, 01 Jan 1970 00:00:00 +0000</created>
            <summary>Bug A</summary>
            <comment>
              <uuid>uuid-commA</uuid>
              <short-name>abc/a/uui</short-name>
              <author></author>
              <date>Thu, 01 Jan 1970 00:01:00 +0000</date>
              <content-type>text/plain</content-type>
              <body>comment A</body>
            </comment>
          </bug>
          <bug>
            <uuid>b</uuid>
            <short-name>abc/b</short-name>
            <severity>minor</severity>
            <status>closed</status>
            <creator>Jane Doe &lt;jdoe@example.com&gt;</creator>
            <created>Thu, 01 Jan 1970 00:00:00 +0000</created>
            <summary>Bug B</summary>
          </bug>
          <bug>
            <uuid>c</uuid>
            <short-name>abc/c</short-name>
            <severity>minor</severity>
            <status>open</status>
            <created>Thu, 01 Jan 1970 00:02:00 +0000</created>
            <summary>bug C</summary>
          </bug>
        </bugdir>
        >>> bugdirA.cleanup()
        >>> bugdirB.cleanup()
        """
        if hasattr(other, 'explicit_attrs'):
            for attr in other.explicit_attrs:
                old = getattr(self, attr)
                new = getattr(other, attr)
                if old != new:
                    if accept_changes:
                        setattr(self, attr, new)
                    elif change_exception:
                        raise ValueError(
                            ('Merge would change {0} "{1}"->"{2}" '
                             'for bugdir {3}'
                             ).format(attr, old, new, self.uuid))
        for estr in other.extra_strings:
            if not estr in self.extra_strings:
                if accept_extra_strings:
                    self.extra_strings += [estr]
                elif change_exception:
                    raise ValueError(
                        ('Merge would add extra string "{0}" for bugdir {1}'
                         ).format(estr, self.uuid))
        for o_bug in other:
            try:
                s_bug = self.bug_from_uuid(o_bug.uuid)
            except KeyError as e:
                try:
                    s_bug = self.bug_from_uuid(o_bug.alt_id)
                except KeyError as e:
                    s_bug = None
            if s_bug is None:
                if accept_bugs:
                    o_bug_copy = copy.copy(o_bug)
                    o_bug_copy.bugdir = self
                    o_bug_copy.id = libbe.util.id.ID(o_bug_copy, 'bug')
                    self.append(o_bug_copy)
                elif change_exception:
                    raise ValueError(
                        ('Merge would add bug {0} (alt: {1}) to bugdir {2}'
                         ).format(o_bug.uuid, o_bug.alt_id, self.uuid))
            else:
                s_bug.merge(o_bug, accept_changes=accept_changes,
                            accept_extra_strings=accept_extra_strings,
                            change_exception=change_exception)

    # methods for id generation

    def sibling_uuids(self):
        return []

class RevisionedBugDir (BugDir):
    """
    RevisionedBugDirs are read-only copies used for generating
    diffs between revisions.
    """
    def __init__(self, bugdir, revision):
        storage_version = bugdir.storage.storage_version(revision)
        if storage_version != libbe.storage.STORAGE_VERSION:
            raise libbe.storage.InvalidStorageVersion(storage_version)
        s = copy.deepcopy(bugdir.storage)
        s.writeable = False
        class RevisionedStorage (object):
            def __init__(self, storage, default_revision):
                self.s = storage
                self.sget = self.s.get
                self.sancestors = self.s.ancestors
                self.schildren = self.s.children
                self.schanged = self.s.changed
                self.r = default_revision
            def get(self, *args, **kwargs):
                if not 'revision' in kwargs or kwargs['revision'] == None:
                    kwargs['revision'] = self.r
                return self.sget(*args, **kwargs)
            def ancestors(self, *args, **kwargs):
                print 'getting ancestors', args, kwargs
                if not 'revision' in kwargs or kwargs['revision'] == None:
                    kwargs['revision'] = self.r
                ret = self.sancestors(*args, **kwargs)
                print 'got ancestors', ret
                return ret
            def children(self, *args, **kwargs):
                if not 'revision' in kwargs or kwargs['revision'] == None:
                    kwargs['revision'] = self.r
                return self.schildren(*args, **kwargs)
            def changed(self, *args, **kwargs):
                if not 'revision' in kwargs or kwargs['revision'] == None:
                    kwargs['revision'] = self.r
                return self.schanged(*args, **kwargs)
        rs = RevisionedStorage(s, revision)
        s.get = rs.get
        s.ancestors = rs.ancestors
        s.children = rs.children
        s.changed = rs.changed
        BugDir.__init__(self, s, from_storage=True)
        self.revision = revision
    def changed(self):
        return self.storage.changed()
    

if libbe.TESTING == True:
    class SimpleBugDir (BugDir):
        """
        For testing.  Set ``memory=True`` for a memory-only bugdir.

        >>> bugdir = SimpleBugDir()
        >>> uuids = list(bugdir.uuids())
        >>> uuids.sort()
        >>> print uuids
        ['a', 'b']
        >>> bugdir.cleanup()
        """
        def __init__(self, memory=True, versioned=False):
            if memory == True:
                storage = None
            else:
                dir = utility.Dir()
                self._dir_ref = dir # postpone cleanup since dir.cleanup() removes dir.
                if versioned == False:
                    storage = libbe.storage.base.Storage(dir.path)
                else:
                    storage = libbe.storage.base.VersionedStorage(dir.path)
                storage.init()
                storage.connect()
            BugDir.__init__(self, storage=storage, uuid='abc123')
            bug_a = self.new_bug(summary='Bug A', _uuid='a')
            bug_a.creator = 'John Doe <jdoe@example.com>'
            bug_a.time = 0
            bug_b = self.new_bug(summary='Bug B', _uuid='b')
            bug_b.creator = 'Jane Doe <jdoe@example.com>'
            bug_b.time = 0
            bug_b.status = 'closed'
            if self.storage != None:
                self.storage.disconnect() # flush to storage
                self.storage.connect()

        def cleanup(self):
            if self.storage != None:
                self.storage.writeable = True
                self.storage.disconnect()
                self.storage.destroy()
            if hasattr(self, '_dir_ref'):
                self._dir_ref.cleanup()

        def flush_reload(self):
            if self.storage != None:
                self.storage.disconnect()
                self.storage.connect()
                self._clear_bugs()

#    class BugDirTestCase(unittest.TestCase):
#        def setUp(self):
#            self.dir = utility.Dir()
#            self.bugdir = BugDir(self.dir.path, sink_to_existing_root=False,
#                                 allow_storage_init=True)
#            self.storage = self.bugdir.storage
#        def tearDown(self):
#            self.bugdir.cleanup()
#            self.dir.cleanup()
#        def fullPath(self, path):
#            return os.path.join(self.dir.path, path)
#        def assertPathExists(self, path):
#            fullpath = self.fullPath(path)
#            self.failUnless(os.path.exists(fullpath)==True,
#                            "path %s does not exist" % fullpath)
#            self.assertRaises(AlreadyInitialized, BugDir,
#                              self.dir.path, assertNewBugDir=True)
#        def versionTest(self):
#            if self.storage != None and self.storage.versioned == False:
#                return
#            original = self.bugdir.storage.commit("Began versioning")
#            bugA = self.bugdir.bug_from_uuid("a")
#            bugA.status = "fixed"
#            self.bugdir.save()
#            new = self.storage.commit("Fixed bug a")
#            dupdir = self.bugdir.duplicate_bugdir(original)
#            self.failUnless(dupdir.root != self.bugdir.root,
#                            "%s, %s" % (dupdir.root, self.bugdir.root))
#            bugAorig = dupdir.bug_from_uuid("a")
#            self.failUnless(bugA != bugAorig,
#                            "\n%s\n%s" % (bugA.string(), bugAorig.string()))
#            bugAorig.status = "fixed"
#            self.failUnless(bug.cmp_status(bugA, bugAorig)==0,
#                            "%s, %s" % (bugA.status, bugAorig.status))
#            self.failUnless(bug.cmp_severity(bugA, bugAorig)==0,
#                            "%s, %s" % (bugA.severity, bugAorig.severity))
#            self.failUnless(bug.cmp_assigned(bugA, bugAorig)==0,
#                            "%s, %s" % (bugA.assigned, bugAorig.assigned))
#            self.failUnless(bug.cmp_time(bugA, bugAorig)==0,
#                            "%s, %s" % (bugA.time, bugAorig.time))
#            self.failUnless(bug.cmp_creator(bugA, bugAorig)==0,
#                            "%s, %s" % (bugA.creator, bugAorig.creator))
#            self.failUnless(bugA == bugAorig,
#                            "\n%s\n%s" % (bugA.string(), bugAorig.string()))
#            self.bugdir.remove_duplicate_bugdir()
#            self.failUnless(os.path.exists(dupdir.root)==False,
#                            str(dupdir.root))
#        def testRun(self):
#            self.bugdir.new_bug(uuid="a", summary="Ant")
#            self.bugdir.new_bug(uuid="b", summary="Cockroach")
#            self.bugdir.new_bug(uuid="c", summary="Praying mantis")
#            length = len(self.bugdir)
#            self.failUnless(length == 3, "%d != 3 bugs" % length)
#            uuids = list(self.bugdir.uuids())
#            self.failUnless(len(uuids) == 3, "%d != 3 uuids" % len(uuids))
#            self.failUnless(uuids == ["a","b","c"], str(uuids))
#            bugA = self.bugdir.bug_from_uuid("a")
#            bugAprime = self.bugdir.bug_from_shortname("a")
#            self.failUnless(bugA == bugAprime, "%s != %s" % (bugA, bugAprime))
#            self.bugdir.save()
#            self.versionTest()
#        def testComments(self, sync_with_disk=False):
#            if sync_with_disk == True:
#                self.bugdir.set_sync_with_disk(True)
#            self.bugdir.new_bug(uuid="a", summary="Ant")
#            bug = self.bugdir.bug_from_uuid("a")
#            comm = bug.comment_root
#            rep = comm.new_reply("Ants are small.")
#            rep.new_reply("And they have six legs.")
#            if sync_with_disk == False:
#                self.bugdir.save()
#                self.bugdir.set_sync_with_disk(True)
#            self.bugdir._clear_bugs()
#            bug = self.bugdir.bug_from_uuid("a")
#            bug.load_comments()
#            if sync_with_disk == False:
#                self.bugdir.set_sync_with_disk(False)
#            self.failUnless(len(bug.comment_root)==1, len(bug.comment_root))
#            for index,comment in enumerate(bug.comments()):
#                if index == 0:
#                    repLoaded = comment
#                    self.failUnless(repLoaded.uuid == rep.uuid, repLoaded.uuid)
#                    self.failUnless(comment.sync_with_disk == sync_with_disk,
#                                    comment.sync_with_disk)
#                    self.failUnless(comment.content_type == "text/plain",
#                                    comment.content_type)
#                    self.failUnless(repLoaded.settings["Content-type"] == \
#                                        "text/plain",
#                                    repLoaded.settings)
#                    self.failUnless(repLoaded.body == "Ants are small.",
#                                    repLoaded.body)
#                elif index == 1:
#                    self.failUnless(comment.in_reply_to == repLoaded.uuid,
#                                    repLoaded.uuid)
#                    self.failUnless(comment.body == "And they have six legs.",
#                                    comment.body)
#                else:
#                    self.failIf(True,
#                                "Invalid comment: %d\n%s" % (index, comment))
#        def testSyncedComments(self):
#            self.testComments(sync_with_disk=True)

    class SimpleBugDirTestCase (unittest.TestCase):
        def setUp(self):
            # create a pre-existing bugdir in a temporary directory
            self.dir = utility.Dir()
            self.storage = libbe.storage.base.Storage(self.dir.path)
            self.storage.init()
            self.storage.connect()
            self.bugdir = BugDir(self.storage)
            self.bugdir.new_bug(summary="Hopefully not imported",
                                _uuid="preexisting")
            self.storage.disconnect()
            self.storage.connect()
        def tearDown(self):
            if self.storage != None:
                self.storage.disconnect()
                self.storage.destroy()
            self.dir.cleanup()
        def testOnDiskCleanLoad(self):
            """
            SimpleBugDir(memory==False) should not import
            preexisting bugs.
            """
            bugdir = SimpleBugDir(memory=False)
            self.failUnless(bugdir.storage.is_readable() == True,
                            bugdir.storage.is_readable())
            self.failUnless(bugdir.storage.is_writeable() == True,
                            bugdir.storage.is_writeable())
            uuids = sorted([bug.uuid for bug in bugdir])
            self.failUnless(uuids == ['a', 'b'], uuids)
            bugdir.flush_reload()
            uuids = sorted(bugdir.uuids())
            self.failUnless(uuids == ['a', 'b'], uuids)
            uuids = sorted([bug.uuid for bug in bugdir])
            self.failUnless(uuids == [], uuids)
            bugdir.load_all_bugs()
            uuids = sorted([bug.uuid for bug in bugdir])
            self.failUnless(uuids == ['a', 'b'], uuids)
            bugdir.cleanup()
        def testInMemoryCleanLoad(self):
            """
            SimpleBugDir(memory==True) should not import
            preexisting bugs.
            """
            bugdir = SimpleBugDir(memory=True)
            self.failUnless(bugdir.storage == None, bugdir.storage)
            uuids = sorted([bug.uuid for bug in bugdir])
            self.failUnless(uuids == ['a', 'b'], uuids)
            uuids = sorted([bug.uuid for bug in bugdir])
            self.failUnless(uuids == ['a', 'b'], uuids)
            bugdir._clear_bugs()
            uuids = sorted(bugdir.uuids())
            self.failUnless(uuids == [], uuids)
            uuids = sorted([bug.uuid for bug in bugdir])
            self.failUnless(uuids == [], uuids)
            bugdir.cleanup()

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])

#    def _get_settings(self, settings_path, for_duplicate_bugdir=False):
#        allow_no_storage = not self.storage.path_in_root(settings_path)
#        if allow_no_storage == True:
#            assert for_duplicate_bugdir == True
#        if self.sync_with_disk == False and for_duplicate_bugdir == False:
#            # duplicates can ignore this bugdir's .sync_with_disk status
#            raise DiskAccessRequired("_get settings")
#        try:
#            settings = mapfile.map_load(self.storage, settings_path, allow_no_storage)
#        except storage.NoSuchFile:
#            settings = {"storage_name": "None"}
#        return settings

#    def _save_settings(self, settings_path, settings,
#                       for_duplicate_bugdir=False):
#        allow_no_storage = not self.storage.path_in_root(settings_path)
#        if allow_no_storage == True:
#            assert for_duplicate_bugdir == True
#        if self.sync_with_disk == False and for_duplicate_bugdir == False:
#            # duplicates can ignore this bugdir's .sync_with_disk status
#            raise DiskAccessRequired("_save settings")
#        self.storage.mkdir(self.get_path(), allow_no_storage)
#        mapfile.map_save(self.storage, settings_path, settings, allow_no_storage)



# Copyright (C) 2010-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for building WSGI commands.

See Also
--------
:py:mod:`libbe.command.serve_storage` and
:py:mod:`libbe.command.serve_commands`.
"""

import copy
import hashlib
import logging
import logging.handlers
import os
import os.path
import re
import select
import signal
import StringIO
import sys
import time
import traceback
import types
import urllib
import urlparse
import wsgiref.simple_server

try:
    import cherrypy
    import cherrypy.wsgiserver
except ImportError:
    cherrypy = None
if cherrypy != None:
    try: # CherryPy >= 3.2
        import cherrypy.wsgiserver.ssl_builtin
    except ImportError: # CherryPy <= 3.1.X
        cherrypy.wsgiserver.ssl_builtin = None

try:
    import OpenSSL
except ImportError:
    OpenSSL = None

import libbe.command
import libbe.command.base
import libbe.command.util
import libbe.storage
import libbe.util.encoding
import libbe.util.http
import libbe.util.id


if libbe.TESTING == True:
    import doctest
    import unittest
    import wsgiref.validate
    try:
        import cherrypy.test.webtest
        cherrypy_test_webtest = True
    except ImportError:
        cherrypy_test_webtest = None


class HandlerError (Exception):
    def __init__(self, code, msg, headers=[]):
        super(HandlerError, self).__init__('{0} {1}'.format(code, msg))
        self.code = code
        self.msg = msg
        self.headers = headers


class Unauthenticated (HandlerError):
    def __init__(self, realm, msg='User Not Authenticated', headers=[]):
        super(Unauthenticated, self).__init__(401, msg, headers+[
                ('WWW-Authenticate','Basic realm="{0}"'.format(realm))])


class Unauthorized (HandlerError):
    def __init__(self, msg='User Not Authorized', headers=[]):
        super(Unauthorized, self).__init__(403, msg, headers)


class User (object):
    def __init__(self, uname=None, name=None, passhash=None, password=None):
        self.uname = uname
        self.name = name
        self.passhash = passhash
        if passhash is None:
            if password is not None:
                self.passhash = self.hash(password)
        else:
            assert password is None, (
                'Redundant password {0} with passhash {1}'.format(
                    password, passhash))
        self.users = None

    def from_string(self, string):
        string = string.strip()
        fields = string.split(':')
        if len(fields) != 3:
            raise ValueError, '{0}!=3 fields in "{1}"'.format(
                len(fields), string)
        self.uname,self.name,self.passhash = fields

    def __str__(self):
        return ':'.join([self.uname, self.name, self.passhash])

    def __cmp__(self, other):
        return cmp(self.uname, other.uname)

    def hash(self, password):
        return hashlib.sha1(password).hexdigest()

    def valid_login(self, password):
        if self.hash(password) == self.passhash:
            return True
        return False

    def set_name(self, name):
        self._set_property('name', name)

    def set_password(self, password):
        self._set_property('passhash', self.hash(password))

    def _set_property(self, property, value):
        if self.uname == 'guest':
            raise Unauthorized(
                'guest user not allowed to change {0}'.format(property))
        if (getattr(self, property) != value and
            self.users is not None):
            self.users.changed = True
        setattr(self, property, value)


class Users (dict):
    def __init__(self, filename=None):
        super(Users, self).__init__()
        self.filename = filename
        self.changed = False

    def load(self):
        if self.filename is None:
            return
        user_file = libbe.util.encoding.get_file_contents(
            self.filename, decode=True)
        self.clear()
        for line in user_file.splitlines():
            user = User()
            user.from_string(line)
            self.add_user(user)

    def save(self):
        if self.filename is not None and self.changed:
            lines = []
            for user in sorted(self.users):
                lines.append(str(user))
            libbe.util.encoding.set_file_contents(self.filename)
            self.changed = False

    def add_user(self, user):
        assert user.users is None, user.users
        user.users = self
        self[user.uname] = user

    def valid_login(self, uname, password):
        return (uname in self and
                self[uname].valid_login(password))


class WSGI_Object (object):
    """Utility class for WGSI clients and middleware.

    For details on WGSI, see `PEP 333`_

    .. _PEP 333: http://www.python.org/dev/peps/pep-0333/
    """
    def __init__(self, logger=None, log_level=logging.INFO, log_format=None):
        self.logger = logger
        self.log_level = log_level
        if log_format is None:
            self.log_format = (
                '{REMOTE_ADDR} - {REMOTE_USER} [{time}] '
                '"{REQUEST_METHOD} {REQUEST_URI} {HTTP_VERSION}" '
                '{status} {bytes} "{HTTP_REFERER}" "{HTTP_USER_AGENT}"')
        else:
            self.log_format = log_format

    def __call__(self, environ, start_response):
        if self.logger is not None:
            self.logger.log(
                logging.DEBUG, 'entering {0}'.format(self.__class__.__name__))
        ret = self._call(environ, start_response)
        if self.logger is not None:
            self.logger.log(
                logging.DEBUG, 'leaving {0}'.format(self.__class__.__name__))
        return ret

    def _call(self, environ, start_response):
        """The main WSGI entry point."""
        raise NotImplementedError
        # start_response() is a callback for setting response headers
        #   start_response(status, response_headers, exc_info=None)
        # status is an HTTP status string (e.g., "200 OK").
        # response_headers is a list of 2-tuples, the HTTP headers in
        # key-value format.
        # exc_info is used in exception handling.
        #
        # The application function then returns an iterable of body chunks.

    def error(self, environ, start_response, error, message, headers=[]):
        """Make it easy to call start_response for errors."""
        response = '{0} {1}'.format(error, message)
        self.log_request(environ, status=response, bytes=len(message))
        start_response(response,
                       [('Content-Type', 'text/plain')]+headers)
        return [message]

    def log_request(self, environ, status='-1 OK', bytes=-1):
        if self.logger is None or self.logger.level > self.log_level:
            return
        req_uri = urllib.quote(environ.get('SCRIPT_NAME', '')
                               + environ.get('PATH_INFO', ''))
        if environ.get('QUERY_STRING'):
            req_uri += '?' + environ['QUERY_STRING']
        start = time.localtime()
        if time.daylight:
            offset = time.altzone / 60 / 60 * -100
        else:
            offset = time.timezone / 60 / 60 * -100
        if offset >= 0:
            offset = '+{0:04d}'.format(offset)
        elif offset < 0:
            offset = '{0:04d}'.format(offset)
        d = {
            'REMOTE_ADDR': environ.get('REMOTE_ADDR', '-'),
            'REMOTE_USER': environ.get('REMOTE_USER', '-'),
            'REQUEST_METHOD': environ['REQUEST_METHOD'],
            'REQUEST_URI': req_uri,
            'HTTP_VERSION': environ.get('SERVER_PROTOCOL'),
            'time': time.strftime('%d/%b/%Y:%H:%M:%S ', start) + offset,
            'status': status.split(None, 1)[0],
            'bytes': bytes,
            'HTTP_REFERER': environ.get('HTTP_REFERER', '-'),
            'HTTP_USER_AGENT': environ.get('HTTP_USER_AGENT', '-'),
            }
        self.logger.log(self.log_level, self.log_format.format(**d))


class WSGI_Middleware (WSGI_Object):
    """Utility class for WGSI middleware.
    """
    def __init__(self, app, *args, **kwargs):
        super(WSGI_Middleware, self).__init__(*args, **kwargs)
        self.app = app

    def _call(self, environ, start_response):
        return self.app(environ, start_response)


class ExceptionApp (WSGI_Middleware):
    """Some servers (e.g. cherrypy) eat app-raised exceptions.

    Work around that by logging tracebacks by hand.
    """
    def _call(self, environ, start_response):
        try:
            return self.app(environ, start_response)
        except Exception, e:
            etype,value,tb = sys.exc_info()
            trace = ''.join(
                traceback.format_exception(etype, value, tb, None))
            self.logger.log(self.log_level, trace)
            raise


class HandlerErrorApp (WSGI_Middleware):
    """Catch HandlerErrors and return HTTP error pages.
    """
    def _call(self, environ, start_response):
        try:
            return self.app(environ, start_response)
        except HandlerError, e:
            self.log_request(environ, status=str(e), bytes=0)
            start_response('{0} {1}'.format(e.code, e.msg), e.headers)
            return []


class BEExceptionApp (WSGI_Middleware):
    """Translate BE-specific exceptions
    """
    def __init__(self, *args, **kwargs):
        super(BEExceptionApp, self).__init__(*args, **kwargs)

    def _call(self, environ, start_response):
        try:
            return self.app(environ, start_response)
        except libbe.storage.NotReadable as e:
            raise libbe.util.wsgi.HandlerError(403, 'Read permission denied')
        except libbe.storage.NotWriteable as e:
            raise libbe.util.wsgi.HandlerError(403, 'Write permission denied')
        except (libbe.command.UsageError,
                libbe.command.UserError,
                OSError,
                libbe.storage.ConnectionError,
                libbe.util.http.HTTPError,
                libbe.util.id.MultipleIDMatches,
                libbe.util.id.NoIDMatches,
                libbe.util.id.InvalidIDStructure,
                libbe.storage.InvalidID,
                ) as e:
            msg = '{0} {1}'.format(type(e).__name__, format(e))
            raise libbe.util.wsgi.HandlerError(
                libbe.util.http.HTTP_USER_ERROR, msg)


class UppercaseHeaderApp (WSGI_Middleware):
    """WSGI middleware that uppercases incoming HTTP headers.

    From PEP 333, `The start_response() Callable`_ :

        A reminder for server/gateway authors: HTTP
        header names are case-insensitive, so be sure
        to take that into consideration when examining
        application-supplied headers!

    .. _The start_response() Callable:
      http://www.python.org/dev/peps/pep-0333/#id20
    """
    def _call(self, environ, start_response):
        for key,value in environ.items():
            if key.startswith('HTTP_'):
                uppercase = key.upper()
                if uppercase != key:
                    environ[uppercase] = environ.pop(key)
        return self.app(environ, start_response)


class AuthenticationApp (WSGI_Middleware):
    """WSGI middleware for handling user authentication.
    """
    def __init__(self, realm, setting='be-auth', users=None, *args, **kwargs):
        super(AuthenticationApp, self).__init__(*args, **kwargs)
        self.realm = realm
        self.setting = setting
        self.users = users

    def _call(self, environ, start_response):
        environ['{0}.realm'.format(self.setting)] = self.realm
        try:
            username = self.authenticate(environ)
            environ['{0}.user'.format(self.setting)] = username
            environ['{0}.user.name'.format(self.setting)
                    ] = self.users[username].name
            return self.app(environ, start_response)
        except Unauthorized, e:
            return self.error(environ, start_response,
                              e.code, e.msg, e.headers)

    def authenticate(self, environ):
        """Handle user-authentication sent in the "Authorization" header.

        This function implements ``Basic`` authentication as described in
        HTTP/1.0 specification [1]_ .  Do not use this module unless you
        are using SSL, as it transmits unencrypted passwords.

        .. [1] http://www.w3.org/Protocols/HTTP/1.0/draft-ietf-http-spec.html#BasicAA

        Examples
        --------

        >>> users = Users()
        >>> users.add_user(User('Aladdin', 'Big Al', password='open sesame'))
        >>> app = AuthenticationApp(app=None, realm='Dummy Realm', users=users)
        >>> app.authenticate({'HTTP_AUTHORIZATION':'Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ=='})
        'Aladdin'
        >>> app.authenticate({'HTTP_AUTHORIZATION':'Basic AAAAAAAAAAAAAAAAAAAAAAAAAA=='})

        Notes
        -----

        Code based on authkit/authenticate/basic.py
        (c) 2005 Clark C. Evans.
        Released under the MIT License:
        http://www.opensource.org/licenses/mit-license.php
        """
        authorization = environ.get('HTTP_AUTHORIZATION', None)
        if authorization is None:
            raise Unauthorized('Authorization required')
        try:
            authmeth,auth = authorization.split(' ', 1)
        except ValueError:
            return None
        if 'basic' != authmeth.lower():
            return None  # non-basic HTTP authorization not implemented
        auth = auth.strip().decode('base64')
        try:
            username,password = auth.split(':', 1)
        except ValueError:
            return None
        if self.authfunc(environ, username, password):
            return username

    def authfunc(self, environ, username, password):
        if not username in self.users:
            return False
        if self.users[username].valid_login(password):
            if self.logger is not None:
                self.logger.log(self.log_level,
                    'Authenticated {0}'.format(self.users[username].name))
            return True
        return False


class WSGI_DataObject (WSGI_Object):
    """Useful WSGI utilities for handling data (POST, QUERY) and
    returning responses.
    """
    def __init__(self, *args, **kwargs):
        super(WSGI_DataObject, self).__init__(*args, **kwargs)

        # Maximum input we will accept when REQUEST_METHOD is POST
        # 0 ==> unlimited input
        self.maxlen = 0

    def ok_response(self, environ, start_response, content,
                    content_type='application/octet-stream',
                    headers=[]):
        if content is None:
            start_response('200 OK', [])
            return []
        if type(content) is types.UnicodeType:
            content = content.encode('utf-8')
        for i,header in enumerate(headers):
            header_name,header_value = header
            if type(header_value) == types.UnicodeType:
                headers[i] = (header_name, header_value.encode('ISO-8859-1'))
        response = '200 OK'
        content_length = len(content)
        self.log_request(environ, status=response, bytes=content_length)
        start_response(response, [
                ('Content-Type', content_type),
                ('Content-Length', str(content_length)),
                ]+headers)
        if self.is_head(environ):
            return []
        return [content]

    def query_data(self, environ):
        if not environ['REQUEST_METHOD'] in ['GET', 'HEAD']:
            raise HandlerError(404, 'Not Found')
        return self._parse_query(environ.get('QUERY_STRING', ''))

    def _parse_query(self, query):
        if len(query) == 0:
            return {}
        data = urlparse.parse_qs(
            query, keep_blank_values=True, strict_parsing=True)
        for k,v in data.items():
            if len(v) == 1:
                data[k] = v[0]
        return data

    def post_data(self, environ):
        if environ['REQUEST_METHOD'] != 'POST':
            raise HandlerError(404, 'Not Found')
        post_data = self._read_post_data(environ)
        return self._parse_post(post_data)

    def _parse_post(self, post):
        return self._parse_query(post)

    def _read_post_data(self, environ):
        try:
            clen = int(environ.get('CONTENT_LENGTH', '0'))
        except ValueError:
            clen = 0
        if clen != 0:
            if self.maxlen > 0 and clen > self.maxlen:
                raise ValueError, 'Maximum content length exceeded'
            return environ['wsgi.input'].read(clen)
        return ''

    def data_get_string(self, data, key, default=None, source='query'):
        if not key in data or data[key] in [None, 'None']:
            if default == HandlerError:
                raise HandlerError(
                    406, 'Missing {0} key {1}'.format(source, key))
            return default
        return data[key]

    def data_get_id(self, data, key='id', default=HandlerError,
                    source='query'):
        return self.data_get_string(data, key, default, source)

    def data_get_boolean(self, data, key, default=False, source='query'):
        val = self.data_get_string(data, key, default, source)
        if val == 'True':
            return True
        elif val == 'False':
            return False
        return val

    def is_head(self, environ):
        return environ['REQUEST_METHOD'] == 'HEAD'


class WSGI_AppObject (WSGI_Object):
    """Useful WSGI utilities for handling URL delegation.
    """
    def __init__(self, urls=tuple(), default_handler=None, setting='be-server',
                 *args, **kwargs):
        super(WSGI_AppObject, self).__init__(*args, **kwargs)
        self.urls = [(re.compile(regexp),callback) for regexp,callback in urls]
        self.default_handler = default_handler
        self.setting = setting

    def _call(self, environ, start_response):
        path = environ.get('PATH_INFO', '').lstrip('/')
        for regexp,callback in self.urls:
            match = regexp.match(path)
            if match is not None:
                setting = '{0}.url_args'.format(self.setting)
                environ[setting] = match.groups()
                return callback(environ, start_response)
        if self.default_handler is None:
            raise HandlerError(404, 'Not Found')
        return self.default_handler(environ, start_response)


class AdminApp (WSGI_AppObject, WSGI_DataObject, WSGI_Middleware):
    """WSGI middleware for managing users

    Changing passwords, usernames, etc.
    """
    def __init__(self, users=None, setting='be-auth', *args, **kwargs):
        handler = ('^admin/?', self.admin)
        if 'urls' not in kwargs:
            kwargs['urls'] = [handler]
        else:
            kwargs.urls.append(handler)
        super(AdminApp, self).__init__(*args, **kwargs)
        self.users = users
        self.setting = setting

    def admin(self, environ, start_response):
        if not '{0}.user'.format(self.setting) in environ:
            realm = envirion.get('{0}.realm'.format(self.setting))
            raise Unauthenticated(realm=realm)
        uname = environ.get('{0}.user'.format(self.setting))
        user = self.users[uname]
        data = self.post_data(environ)
        source = 'post'
        name = self.data_get_string(
            data, 'name', default=None, source=source)
        if name is not None:
            self.users[uname].set_name(name)
        password = self.data_get_string(
            data, 'password', default=None, source=source)
        if password is not None:
            self.users[uname].set_password(password)
        self.users.save()
        return self.ok_response(environ, start_response, None)


class SilentRequestHandler (wsgiref.simple_server.WSGIRequestHandler):
    def log_message(self, format, *args):
        pass


class ServerCommand (libbe.command.base.Command):
    """Serve something over HTTP.

    Use this as a base class to build commands that serve a web interface.
    """
    _daemon_actions = ['start', 'stop']
    _daemon_action_present_participle = {
        'start': 'starting',
        'stop': 'stopping',
        }

    def __init__(self, *args, **kwargs):
        super(ServerCommand, self).__init__(*args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='port',
                    help='Bind server to port',
                    arg=libbe.command.Argument(
                        name='port', metavar='INT', type='int', default=8000)),
                libbe.command.Option(name='host',
                    help='Set host string (blank for localhost)',
                    arg=libbe.command.Argument(
                        name='host', metavar='HOST', default='localhost')),
                libbe.command.Option(name='daemon',
                    help=('Start or stop a server daemon.  Stopping requires '
                          'a PID file'),
                    arg=libbe.command.Argument(
                        name='daemon', metavar='ACTION',
                        completion_callback=libbe.command.util.Completer(
                            self._daemon_actions))),
                libbe.command.Option(name='pidfile', short_name='p',
                    help='Store the process id in the given path',
                    arg=libbe.command.Argument(
                        name='pidfile', metavar='FILE',
                        completion_callback=libbe.command.util.complete_path)),
                libbe.command.Option(name='logfile',
                    help='Log to the given path (instead of stdout)',
                    arg=libbe.command.Argument(
                        name='logfile', metavar='FILE',
                        completion_callback=libbe.command.util.complete_path)),
                libbe.command.Option(name='read-only', short_name='r',
                    help='Dissable operations that require writing'),
                libbe.command.Option(name='notify', short_name='n',
                    help='Send notification emails for changes.',
                    arg=libbe.command.Argument(
                        name='notify', metavar='EMAIL-COMMAND', default=None)),
                libbe.command.Option(name='ssl', short_name='s',
                    help='Use CherryPy to serve HTTPS (HTTP over SSL/TLS)'),
                libbe.command.Option(name='auth', short_name='a',
                    help=('Require authentication.  FILE should be a file '
                          'containing colon-separated '
                          'UNAME:USER:sha1(PASSWORD) lines, for example: '
                          '"jdoe:John Doe <jdoe@example.com>:'
                          'd99f8e5a4b02dc25f49da2ea67c0034f61779e72"'),
                    arg=libbe.command.Argument(
                        name='auth', metavar='FILE', default=None,
                        completion_callback=libbe.command.util.complete_path)),
                ])

    def _run(self, **params):
        if params['daemon'] not in self._daemon_actions + [None]:
            raise libbe.command.UserError(
                'Invalid daemon action "{0}".\nValid actions:\n  {1}'.format(
                    params['daemon'], self._daemon_actions))
        self._setup_logging(params)
        if params['daemon'] not in [None, 'start']:
            self._manage_daemon(params)
            return
        storage = self._get_storage()
        if params['read-only']:
            writeable = storage.writeable
            storage.writeable = False
        if params['auth']:
            self._check_restricted_access(storage, params['auth'])
        users = Users(params['auth'])
        users.load()
        app = self._get_app(logger=self.logger, storage=storage, **params)
        if params['auth']:
            app = AdminApp(app, users=users, logger=self.logger)
            app = AuthenticationApp(app, realm=storage.repo,
                                    users=users, logger=self.logger)
        app = UppercaseHeaderApp(app, logger=self.logger)
        server,details = self._get_server(params, app)
        details['repo'] = storage.repo
        try:
            self._start_server(params, server, details)
        except KeyboardInterrupt:
            pass
        self._stop_server(params, server)
        if params['read-only']:
            storage.writeable = writeable

    def _get_app(self, logger, storage, **kwargs):
        raise NotImplementedError()

    def _setup_logging(self, params, log_level=logging.INFO):
        self.logger = logging.getLogger('be.{0}'.format(self.name))
        self.log_level = log_level
        if params['logfile']:
            path = os.path.abspath(os.path.expanduser(
                    params['logfile']))
            handler = logging.handlers.TimedRotatingFileHandler(
                path, when='w6', interval=1, backupCount=4,
                encoding=libbe.util.encoding.get_text_file_encoding())
        else:
            handler = logging.StreamHandler(self.stdout)
        handler.setFormatter(logging.Formatter('%(message)s'))
        self.logger.addHandler(handler)
        self.logger.propagate = False
        if log_level is not None:
            handler.setLevel(log_level)
            self.logger.setLevel(log_level)

    def _get_server(self, params, app):
        details = {
            'socket-name':params['host'],
            'port':params['port'],
            }
        if params['ssl']:
            details['protocol'] = 'HTTPS'
        else:
            details['protocol'] = 'HTTP'
        app = BEExceptionApp(app, logger=self.logger)
        app = HandlerErrorApp(app, logger=self.logger)
        app = ExceptionApp(app, logger=self.logger)
        if params['ssl']:
            if cherrypy is None:
                raise libbe.command.UserError(
                    '--ssl requires the cherrypy module')
            server = cherrypy.wsgiserver.CherryPyWSGIServer(
                (params['host'], params['port']), app)
            #server.throw_errors = True
            #server.show_tracebacks = True
            private_key,certificate = _get_cert_filenames(
                'be-server', logger=self.logger, level=self.log_level)
            if cherrypy.wsgiserver.ssl_builtin is None:
                server.ssl_module = 'builtin'
                server.ssl_private_key = private_key
                server.ssl_certificate = certificate
            else:
                server.ssl_adapter = (
                    cherrypy.wsgiserver.ssl_builtin.BuiltinSSLAdapter(
                        certificate=certificate, private_key=private_key))
        else:
            server = wsgiref.simple_server.make_server(
                params['host'], params['port'], app,
                handler_class=SilentRequestHandler)
        return (server, details)

    def _daemonize(self, params):
        signal.signal(signal.SIGTERM, self._sigterm)
        self.logger.log(self.log_level, 'Daemonizing')
        pid = os.fork()
        if pid > 0:
            os._exit(0)
        os.setsid()
        pid = os.fork()
        if pid > 0:
            os._exit(0)
        self.logger.log(
            self.log_level, 'Daemonized with PID {0}'.format(os.getpid()))

    def _get_pidfile(self, params):
        params['pidfile'] = os.path.abspath(os.path.expanduser(
                params['pidfile']))
        self.logger.log(
            self.log_level, 'Get PID file at {0}'.format(params['pidfile']))
        if os.path.exists(params['pidfile']):
            raise libbe.command.UserError(
                'PID file {0} already exists'.format(params['pidfile']))
        pid = os.getpid()
        with open(params['pidfile'], 'w') as f:  # race between exist and open
            f.write(str(os.getpid()))            
        self.logger.log(
            self.log_level, 'Got PID file as {0}'.format(pid))

    def _start_server(self, params, server, details):
        if params['daemon']:
            self._daemonize(params=params)
        if params['pidfile']:
            self._get_pidfile(params)
        self.logger.log(
            self.log_level,
            ('Serving {protocol} on {socket-name} port {port} ...\n'
             'BE repository {repo}').format(**details))
        params['server stopped'] = False
        if isinstance(server, wsgiref.simple_server.WSGIServer):
            try:
                server.serve_forever()
            except select.error as e:
                if len(e.args) == 2 and e.args[1] == 'Interrupted system call':
                    pass
                else:
                    raise
        else:  # CherryPy server
            server.start()

    def _stop_server(self, params, server):
        if params['server stopped']:
            return  # already stopped, e.g. via _sigterm()
        params['server stopped'] = True
        self.logger.log(self.log_level, 'Closing server')
        if isinstance(server, wsgiref.simple_server.WSGIServer):
            server.server_close()
        else:
            server.stop()
        if params['pidfile']:
            os.remove(params['pidfile'])

    def _sigterm(self, signum, frame):
        self.logger.log(self.log_level, 'Handling SIGTERM')
        # extract params and server from the stack
        f = frame
        while f is not None and f.f_code.co_name != '_start_server':
            f = f.f_back
        if f is None:
            self.logger.log(
                self.log_level,
                'SIGTERM from outside _start_server(): {0}'.format(
                    frame.f_code))
            return  # where did this signal come from?
        params = f.f_locals['params']
        server = f.f_locals['server']
        self._stop_server(params=params, server=server)

    def _manage_daemon(self, params):
        "Daemon management (any action besides 'start')"
        if not params['pidfile']:
            raise libbe.command.UserError(
                'daemon management requires --pidfile')
        try:
            with open(params['pidfile'], 'r') as f:
                pid = f.read().strip()
        except IOError as e:
            raise libbe.command.UserError(
                'could not find PID file: {0}'.format(e))
        pid = int(pid)
        pp = self._daemon_action_present_participle[params['daemon']].title()
        self.logger.log(
            self.log_level,
            '{0} daemon running on process {1}'.format(pp, pid))
        if params['daemon'] == 'stop':
            os.kill(pid, signal.SIGTERM)
        else:
            raise NotImplementedError(params['daemon'])

    def _long_help(self):
        raise NotImplementedError()


class WSGICaller (object):
    """Call into WSGI apps programmatically
    """
    def __init__(self, *args, **kwargs):
        super(WSGICaller, self).__init__(*args, **kwargs)
        self.default_environ = { # required by PEP 333
            'REQUEST_METHOD': 'GET', # 'POST', 'HEAD'
            'REMOTE_ADDR': '192.168.0.123',
            'SCRIPT_NAME':'',
            'PATH_INFO': '',
            #'QUERY_STRING':'',   # may be empty or absent
            #'CONTENT_TYPE':'',   # may be empty or absent
            #'CONTENT_LENGTH':'', # may be empty or absent
            'SERVER_NAME':'example.com',
            'SERVER_PORT':'80',
            'SERVER_PROTOCOL':'HTTP/1.1',
            'wsgi.version':(1,0),
            'wsgi.url_scheme':'http',
            'wsgi.input':StringIO.StringIO(),
            'wsgi.errors':StringIO.StringIO(),
            'wsgi.multithread':False,
            'wsgi.multiprocess':False,
            'wsgi.run_once':False,
            }

    def getURL(self, app, path='/', method='GET', data=None,
               data_dict=None, scheme='http', environ={}):
        env = copy.copy(self.default_environ)
        env['PATH_INFO'] = path
        env['REQUEST_METHOD'] = method
        env['scheme'] = scheme
        if data_dict is not None:
            assert data is None, (data, data_dict)
            data = urllib.urlencode(data_dict)
        if data is not None:
            if data_dict is None:
                assert method == 'POST', (method, data)
            if method == 'POST':
                env['CONTENT_LENGTH'] = len(data)
                env['wsgi.input'] = StringIO.StringIO(data)
            else:
                assert method in ['GET', 'HEAD'], method
                env['QUERY_STRING'] = data
        for key,value in environ.items():
            env[key] = value
        return ''.join(app(env, self.start_response))

    def start_response(self, status, response_headers, exc_info=None):
        self.status = status
        self.response_headers = response_headers
        self.exc_info = exc_info


if libbe.TESTING:
    class WSGITestCase (unittest.TestCase):
        def setUp(self):
            self.logstream = StringIO.StringIO()
            self.logger = logging.getLogger('be-wsgi-test')
            console = logging.StreamHandler(self.logstream)
            console.setFormatter(logging.Formatter('%(message)s'))
            self.logger.addHandler(console)
            self.logger.propagate = False
            console.setLevel(logging.INFO)
            self.logger.setLevel(logging.INFO)
            self.caller = WSGICaller()

        def getURL(self, *args, **kwargs):
            content = self.caller.getURL(*args, **kwargs)
            self.status = self.caller.status
            self.response_headers = self.caller.response_headers
            self.exc_info = self.caller.exc_info
            return content

    class WSGI_ObjectTestCase (WSGITestCase):
        def setUp(self):
            WSGITestCase.setUp(self)
            self.app = WSGI_Object(self.logger)

        def test_error(self):
            contents = self.app.error(
                environ=self.caller.default_environ,
                start_response=self.caller.start_response,
                error=123,
                message='Dummy Error',
                headers=[('X-Dummy-Header','Dummy Value')])
            self.failUnless(contents == ['Dummy Error'], contents)
            self.failUnless(
                self.caller.status == '123 Dummy Error', self.caller.status)
            self.failUnless(self.caller.response_headers == [
                    ('Content-Type','text/plain'),
                    ('X-Dummy-Header','Dummy Value')],
                            self.caller.response_headers)
            self.failUnless(self.caller.exc_info == None, self.caller.exc_info)

        def test_log_request(self):
            self.app.log_request(
                environ=self.caller.default_environ, status='-1 OK', bytes=123)
            log = self.logstream.getvalue()
            self.failUnless(log.startswith('192.168.0.123 -'), log)


    class ExceptionAppTestCase (WSGITestCase):
        def setUp(self):
            WSGITestCase.setUp(self)
            def child_app(environ, start_response):
                raise ValueError('Dummy Error')
            self.app = ExceptionApp(child_app, self.logger)

        def test_traceback(self):
            try:
                self.getURL(self.app)
            except ValueError, e:
                pass
            log = self.logstream.getvalue()
            self.failUnless(log.startswith('Traceback'), log)
            self.failUnless('child_app' in log, log)
            self.failUnless('ValueError: Dummy Error' in log, log)


    class AdminAppTestCase (WSGITestCase):
        def setUp(self):
            WSGITestCase.setUp(self)
            self.users = Users()
            self.users.add_user(
                User('Aladdin', 'Big Al', password='open sesame'))
            self.users.add_user(
                User('guest', 'Guest', password='guestpass'))
            def child_app(environ, start_response):
                pass
            app = AdminApp(
                app=child_app, users=self.users, logger=self.logger)
            app = AuthenticationApp(
                app=app, realm='Dummy Realm', users=self.users,
                logger=self.logger)
            self.app = UppercaseHeaderApp(app=app, logger=self.logger)

        def basic_auth(self, uname, password):
            """HTTP basic authorization string"""
            return 'Basic {0}'.format(
                '{0}:{1}'.format(uname, password).encode('base64'))

        def test_new_name(self):
            self.getURL(
                self.app, '/admin/', method='POST',
                data_dict={'name':'Prince Al'},
                environ={'HTTP_Authorization':
                             self.basic_auth('Aladdin', 'open sesame')})
            self.failUnless(self.status == '200 OK', self.status)
            self.failUnless(self.response_headers == [],
                            self.response_headers)
            self.failUnless(self.exc_info == None, self.exc_info)
            self.failUnless(self.users['Aladdin'].name == 'Prince Al',
                            self.users['Aladdin'].name)
            self.failUnless(self.users.changed == True,
                            self.users.changed)

        def test_new_password(self):
            self.getURL(
                self.app, '/admin/', method='POST',
                data_dict={'password':'New Pass'},
                environ={'HTTP_Authorization':
                             self.basic_auth('Aladdin', 'open sesame')})
            self.failUnless(self.status == '200 OK', self.status)
            self.failUnless(self.response_headers == [],
                            self.response_headers)
            self.failUnless(self.exc_info == None, self.exc_info)
            self.failUnless((self.users['Aladdin'].passhash ==
                             self.users['Aladdin'].hash('New Pass')),
                            self.users['Aladdin'].passhash)
            self.failUnless(self.users.changed == True,
                            self.users.changed)

        def test_guest_name(self):
            self.getURL(
                self.app, '/admin/', method='POST',
                data_dict={'name':'SPAM'},
                environ={'HTTP_Authorization':
                             self.basic_auth('guest', 'guestpass')})
            self.failUnless(self.status.startswith('403 '), self.status)
            self.failUnless(self.response_headers == [
                    ('Content-Type', 'text/plain')],
                            self.response_headers)
            self.failUnless(self.exc_info == None, self.exc_info)
            self.failUnless(self.users['guest'].name == 'Guest',
                            self.users['guest'].name)
            self.failUnless(self.users.changed == False,
                            self.users.changed)

        def test_guest_password(self):
            self.getURL(
                self.app, '/admin/', method='POST',
                data_dict={'password':'SPAM'},
                environ={'HTTP_Authorization':
                             self.basic_auth('guest', 'guestpass')})
            self.failUnless(self.status.startswith('403 '), self.status)
            self.failUnless(self.response_headers == [
                    ('Content-Type', 'text/plain')],
                            self.response_headers)
            self.failUnless(self.exc_info == None, self.exc_info)
            self.failUnless(self.users['guest'].name == 'Guest',
                            self.users['guest'].name)
            self.failUnless(self.users.changed == False,
                            self.users.changed)

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])


# The following certificate-creation code is adapted from pyOpenSSL's
# examples.

def _get_cert_filenames(server_name, autogenerate=True, logger=None,
                        level=None):
    """
    Generate private key and certification filenames.
    get_cert_filenames(server_name) -> (pkey_filename, cert_filename)
    """
    pkey_file = '{0}.pkey'.format(server_name)
    cert_file = '{0}.cert'.format(server_name)
    if autogenerate:
        for file in [pkey_file, cert_file]:
            if not os.path.exists(file):
                _make_certs(server_name, logger=logger, level=level)
    return (pkey_file, cert_file)

def _create_key_pair(type, bits):
    """Create a public/private key pair.

    Returns the public/private key pair in a PKey object.

    Parameters
    ----------
    type : TYPE_RSA or TYPE_DSA
      Key type.
    bits : int
      Number of bits to use in the key.
    """
    pkey = OpenSSL.crypto.PKey()
    pkey.generate_key(type, bits)
    return pkey

def _create_cert_request(pkey, digest="md5", **name):
    """Create a certificate request.

    Returns the certificate request in an X509Req object.

    Parameters
    ----------
    pkey : PKey
      The key to associate with the request.
    digest : "md5" or ?
      Digestion method to use for signing, default is "md5",
    `**name` :
      The name of the subject of the request, possible.
      Arguments are:

      ============ ========================
      C            Country name
      ST           State or province name
      L            Locality name
      O            Organization name
      OU           Organizational unit name
      CN           Common name
      emailAddress E-mail address
      ============ ========================
    """
    req = OpenSSL.crypto.X509Req()
    subj = req.get_subject()

    for (key,value) in name.items():
        setattr(subj, key, value)

    req.set_pubkey(pkey)
    req.sign(pkey, digest)
    return req

def _create_certificate(req, (issuerCert, issuerKey), serial,
                        (notBefore, notAfter), digest='md5'):
    """Generate a certificate given a certificate request.

    Returns the signed certificate in an X509 object.

    Parameters
    ----------
    req :
      Certificate reqeust to use
    issuerCert :
      The certificate of the issuer
    issuerKey :
      The private key of the issuer
    serial :
      Serial number for the certificate
    notBefore :
      Timestamp (relative to now) when the certificate
      starts being valid
    notAfter :
      Timestamp (relative to now) when the certificate
      stops being valid
    digest :
      Digest method to use for signing, default is md5
    """
    cert = OpenSSL.crypto.X509()
    cert.set_serial_number(serial)
    cert.gmtime_adj_notBefore(notBefore)
    cert.gmtime_adj_notAfter(notAfter)
    cert.set_issuer(issuerCert.get_subject())
    cert.set_subject(req.get_subject())
    cert.set_pubkey(req.get_pubkey())
    cert.sign(issuerKey, digest)
    return cert

def _make_certs(server_name, logger=None, level=None):
    """Generate private key and certification files.

    `mk_certs(server_name) -> (pkey_filename, cert_filename)`
    """
    if OpenSSL == None:
        raise libbe.command.UserError(
            'SSL certificate generation requires the OpenSSL module')
    pkey_file,cert_file = _get_cert_filenames(
        server_name, autogenerate=False)
    if logger != None:
        logger.log(
            level, 'Generating certificates {0} {1}'.format(
                pkey_file, cert_file))
    cakey = _create_key_pair(OpenSSL.crypto.TYPE_RSA, 1024)
    careq = _create_cert_request(cakey, CN='Certificate Authority')
    cacert = _create_certificate(
        careq, (careq, cakey), 0, (0, 60*60*24*365*5)) # five years
    open(pkey_file, 'w').write(OpenSSL.crypto.dump_privatekey(
            OpenSSL.crypto.FILETYPE_PEM, cakey))
    open(cert_file, 'w').write(OpenSSL.crypto.dump_certificate(
            OpenSSL.crypto.FILETYPE_PEM, cacert))



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""
Assorted utility functions that don't fit in anywhere else.
"""

import calendar
import codecs
import os
import shutil
import tempfile
import time
import types
try: # import core module, Python >= 2.5
    from xml.etree import ElementTree
except ImportError: # look for non-core module
    from elementtree import ElementTree

import libbe
if libbe.TESTING == True:
    import doctest

class InvalidXML(ValueError):
    """Invalid XML while parsing for a `*.from_xml()` method.

    Parameters
    ----------
    type : str
        String identifying `*`, e.g. "bug", "comment", ...
    element : :py:class:`ElementTree.Element`
        ElementTree.Element instance which caused the error.
    error : str
        Error description.
    """
    def __init__(self, type, element, error):
        msg = 'Invalid %s xml: %s\n  %s\n' \
            % (type, error, ElementTree.tostring(element))
        ValueError.__init__(self, msg)
        self.type = type
        self.element = element
        self.error = error

def search_parent_directories(path, filename):
    """
    Find the file (or directory) named filename in path or in any
    of path's parents.  For example::

         search_parent_directories("/a/b/c", ".be")

    will return the path to the first existing file from::

        /a/b/c/.be
        /a/b/.be
        /a/.be
        /.be

    or `None` if none of those files exist.
    """
    path = os.path.realpath(path)
    assert os.path.exists(path)
    old_path = None
    while True:
        check_path = os.path.join(path, filename)
        if os.path.exists(check_path):
            return check_path
        if path == old_path:
            return None
        old_path = path
        path = os.path.dirname(path)

class Dir (object):
    """A temporary directory for testing use.

    Make sure you run :py:meth:`cleanup` after you're done using the
    directory.
    """
    def __init__(self):
        self.path = tempfile.mkdtemp(prefix="BEtest")
        self.removed = False
    def cleanup(self):
        if self.removed == False:
            shutil.rmtree(self.path)
            self.removed = True
    def __call__(self):
        return self.path

RFC_2822_TIME_FMT = "%a, %d %b %Y %H:%M:%S +0000"
"""RFC 2822 [#]_ format string for :py:func:`time.strftime` and
:py:func:`time.strptime`.

.. [#] See `RFC 2822`_, sections 3.3 and A.1.1.
.. _RFC 2822: http://www.faqs.org/rfcs/rfc2822.html
"""

def time_to_str(time_val):
    """Convert a time number into an RFC 2822-formatted string.

    Parameters
    ----------
    time_val : float
      Float seconds since the Epoc, see :py:func:`time.time`.
      Note that while `time_val` may contain sub-second data,
      the output string will not.

    Examples
    --------

    >>> time_to_str(0)
    'Thu, 01 Jan 1970 00:00:00 +0000'

    See Also
    --------
    str_to_time : inverse
    handy_time : localtime string
    """
    return time.strftime(RFC_2822_TIME_FMT, time.gmtime(time_val))

def str_to_time(str_time):
    """Convert an RFC 2822-fomatted string into a time value.

    Parameters
    ----------
    str_time : str
      An RFC 2822-formatted string.

    Examples
    --------

    >>> str_to_time("Thu, 01 Jan 1970 00:00:00 +0000")
    0
    >>> q = time.time()
    >>> str_to_time(time_to_str(q)) == int(q)
    True
    >>> str_to_time("Thu, 01 Jan 1970 00:00:00 -1000")
    36000

    See Also
    --------
    time_to_str : inverse
    """
    timezone_str = str_time[-5:]
    if timezone_str != "+0000":
        str_time = str_time.replace(timezone_str, "+0000")
    time_val = calendar.timegm(time.strptime(str_time, RFC_2822_TIME_FMT))
    timesign = -int(timezone_str[0]+"1") # "+" -> time_val ahead of GMT
    timezone_tuple = time.strptime(timezone_str[1:], "%H%M")
    timezone = timezone_tuple.tm_hour*3600 + timezone_tuple.tm_min*60
    return time_val + timesign*timezone

def handy_time(time_val):
    """Convert a time number into a useful localtime.

    Where :py:func:`time_to_str` returns GMT +0000, `handy_time` returns
    a string in local time.  This may be more accessible for the user.

    Parameters
    ----------
    time_val : float
      Float seconds since the Epoc, see :py:func:`time.time`.
    """
    return time.strftime("%a, %d %b %Y %H:%M", time.localtime(time_val))

def time_to_gmtime(str_time):
    """Convert an RFC 2822-fomatted string to a GMT string.

    Parameters
    ----------
    str_time : str
      An RFC 2822-formatted string.

    Examples
    --------

    >>> time_to_gmtime("Thu, 01 Jan 1970 00:00:00 -1000")
    'Thu, 01 Jan 1970 10:00:00 +0000'
    """
    time_val = str_to_time(str_time)
    return time_to_str(time_val)

def iterable_full_of_strings(value, alternative=None):
    """Require an iterable full of strings.

    This is useful, for example, in validating `*.extra_strings`.
    See :py:attr:`libbe.bugdir.BugDir.extra_strings`

    Parameters
    ----------
    value : list or None
      The potential list of strings.
    alternative
      Allow a default (e.g. `None`), such that::

        iterable_full_of_strings(value=x, alternative=x) -> True

    Examples
    --------

    >>> iterable_full_of_strings([])
    True
    >>> iterable_full_of_strings(["abc", "def", u"hij"])
    True
    >>> iterable_full_of_strings(["abc", None, u"hij"])
    False
    >>> iterable_full_of_strings(None, alternative=None)
    True
    """
    if value == alternative:
        return True
    elif not hasattr(value, '__iter__'):
        return False
    for x in value:
        if type(x) not in types.StringTypes:
            return False
    return True

def underlined(string, char='='):
    """Produces a version of a string that is underlined.

    Parameters
    ----------
    string : str
      The string to underline
    char : str
      The character to use for the underlining.

    Examples
    --------

    >>> underlined("Underlined String")
    'Underlined String\\n================='
    """
    assert len(char) == 1, char
    return '%s\n%s' % (string, char*len(string))

if libbe.TESTING == True:
    suite = doctest.DocTestSuite()



# Copyright (C) 2008-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import copy
import os

import libbe
import libbe.command
import libbe.command.util


class Merge (libbe.command.Command):
    """Merge duplicate bugs

    >>> import sys
    >>> import libbe.bugdir
    >>> import libbe.comment
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Merge(ui=ui)

    >>> a = bd.bug_from_uuid('a')
    >>> a.comment_root.time = 0
    >>> dummy = a.new_comment('Testing')
    >>> dummy.time = 1
    >>> dummy = dummy.new_reply('Testing...')
    >>> dummy.time = 2
    >>> b = bd.bug_from_uuid('b')
    >>> b.status = 'open'
    >>> b.comment_root.time = 0
    >>> dummy = b.new_comment('1 2')
    >>> dummy.time = 1
    >>> dummy = dummy.new_reply('1 2 3 4')
    >>> dummy.time = 2

    >>> ret = ui.run(cmd, args=['/a', '/b'])
    Merged bugs #abc/a# and #abc/b#
    >>> bd.flush_reload()
    >>> a = bd.bug_from_uuid('a')
    >>> a.load_comments()
    >>> a_comments = sorted([c for c in a.comments()],
    ...                     cmp=libbe.comment.cmp_time)
    >>> mergeA = a_comments[0]
    >>> mergeA.time = 3
    >>> print a.string(show_comments=True)
    ... # doctest: +ELLIPSIS, +REPORT_UDIFF
              ID : a
      Short name : abc/a
        Severity : minor
          Status : open
        Assigned : 
        Reporter : 
         Creator : John Doe <jdoe@example.com>
         Created : ...
    Bug A
    --------- Comment ---------
    Name: abc/a/...
    From: ...
    Date: ...
    <BLANKLINE>
    Testing
      --------- Comment ---------
      Name: abc/a/...
      From: ...
      Date: ...
    <BLANKLINE>
      Testing...
    --------- Comment ---------
    Name: abc/a/...
    From: ...
    Date: ...
    <BLANKLINE>
    Merged from bug #abc/b#
      --------- Comment ---------
      Name: abc/a/...
      From: ...
      Date: ...
    <BLANKLINE>
      1 2
        --------- Comment ---------
        Name: abc/a/...
        From: ...
        Date: ...
    <BLANKLINE>
        1 2 3 4
    >>> b = bd.bug_from_uuid('b')
    >>> b.load_comments()
    >>> b_comments = sorted([c for c in b.comments()],
    ...                     libbe.comment.cmp_time)
    >>> mergeB = b_comments[0]
    >>> mergeB.time = 3
    >>> print b.string(show_comments=True)
    ... # doctest: +ELLIPSIS, +REPORT_UDIFF
              ID : b
      Short name : abc/b
        Severity : minor
          Status : closed
        Assigned : 
        Reporter : 
         Creator : Jane Doe <jdoe@example.com>
         Created : ...
    Bug B
    --------- Comment ---------
    Name: abc/b/...
    From: ...
    Date: ...
    <BLANKLINE>
    1 2
      --------- Comment ---------
      Name: abc/b/...
      From: ...
      Date: ...
    <BLANKLINE>
      1 2 3 4
    --------- Comment ---------
    Name: abc/b/...
    From: ...
    Date: ...
    <BLANKLINE>
    Merged into bug #abc/a#
    >>> print b.status
    closed
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'merge'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.args.extend([
                libbe.command.Argument(
                    name='bug-id', metavar='BUG-ID', default=None,
                    completion_callback=libbe.command.util.complete_bug_id),
                libbe.command.Argument(
                    name='bug-id-to-merge', metavar='BUG-ID', default=None,
                    completion_callback=libbe.command.util.complete_bug_id),
                ])

    def _run(self, **params):
        storage = self._get_storage()
        bugdirs = self._get_bugdirs()
        bugdirA,bugA,comment = (
            libbe.command.util.bugdir_bug_comment_from_user_id(
                bugdirs, params['bug-id']))
        bugA.load_comments()
        bugdirB,bugB,dummy_comment = (
            libbe.command.util.bugdir_bug_comment_from_user_id(
                bugdirs, params['bug-id-to-merge']))
        bugB.load_comments()
        mergeA = bugA.new_comment('Merged from bug #%s#' % bugB.id.long_user())
        newCommTree = copy.deepcopy(bugB.comment_root)
        for comment in newCommTree.traverse(): # all descendant comments
            comment.bug = bugA
            # uuids must be unique in storage
            if comment.alt_id == None:
                comment.storage = None
                comment.alt_id = comment.uuid
                comment.storage = storage
            comment.uuid = libbe.util.id.uuid_gen()
            comment.save() # force onto disk under bugA

        for comment in newCommTree: # just the child comments
            mergeA.add_reply(comment, allow_time_inversion=True)
        bugB.new_comment('Merged into bug #%s#' % bugA.id.long_user())
        bugB.status = 'closed'
        print >> self.stdout, 'Merged bugs #%s# and #%s#' \
            % (bugA.id.user(), bugB.id.user())
        return 0

    def _long_help(self):
        return """
The second bug (B) is merged into the first (A).  This adds merge
comments to both bugs, closes B, and appends B's comment tree to A's
merge comment.
"""
